#!/usr/bin/env python3

import argparse
import os
import sys

import pelita
from .script_utils import start_logging
from pelita.ui.tk_viewer import TkViewer


def geometry_string(s):
    """Get a X-style geometry definition and return a tuple.

    600x400 -> (600,400)
    """
    try:
        x_string, y_string = s.split('x')
        geometry = (int(x_string), int(y_string))
    except ValueError:
        msg = "%s is not a valid geometry specification" %s
        raise argparse.ArgumentTypeError(msg)
    return geometry

LOG_TK = os.environ.get("PELITA_LOG_TK", None)

parser = argparse.ArgumentParser(description='Open a Tk viewer')
parser.add_argument('subscribe_sock', metavar="URL", type=str,
                    help='subscribe socket')
parser.add_argument('--controller-address', metavar="URL", type=str,
                    help='controller address')
parser.add_argument('--geometry', type=geometry_string,
                    help='geometry')
parser.add_argument('--fullscreen', const=True, action='store_const',
                     help='fullscreen')
parser.add_argument('--delay', type=int,
                    help='delay')
parser.add_argument('--stop-after', type=int, metavar="N",
                    help='Stop after N rounds.')
parser._optionals = parser.add_argument_group('Options')
parser.add_argument('--version', help='show the version number and exit',
                    action='store_const', const=True)
parser.add_argument('--log', help='print debugging log information to'
                                  ' LOGFILE (default \'stderr\')',
                    metavar='LOGFILE', const='-', nargs='?')

def main():
    args = parser.parse_args()
    if args.version:
        print("Pelita {}".format(pelita.__version__))
        sys.exit(0)

    if LOG_TK or args.log:
        start_logging(args.log)

    tkargs = {
        'address': args.subscribe_sock,
        'controller_address': args.controller_address,
        'geometry': args.geometry,
        'fullscreen' : args.fullscreen,
        'delay': args.delay,
        'stop_after': args.stop_after
    }
    v = TkViewer(**{k: v for k, v in list(tkargs.items()) if v is not None})
    v.run()

if __name__ == '__main__':
    main()




from .base import (stepping_player, speaking_player,
                   round_based_player, move_exception_player,
                   debuggable_player)


from .RandomPlayers import random_player, nq_random_player
from .FoodEatingPlayer import food_eating_player
from .SmartEatingPlayer import smart_eating_player
from .RandomExplorerPlayer import random_explorer_player
from .SmartRandomPlayer import smart_random_player
from .StoppingPlayer import stopping_player

SANE_PLAYERS = [
    random_player,
    nq_random_player,
    food_eating_player,
    smart_eating_player,
    random_explorer_player,
    smart_random_player]



import logging

def start_logging(filename, module='pelita'):
    if not filename or filename == '-':
        hdlr = logging.StreamHandler()
    else:
        hdlr = logging.FileHandler(filename, mode='w')
    logger = logging.getLogger(module)
    FORMAT = '[%(relativeCreated)06d %(name)s:%(levelname).1s][%(funcName)s] %(message)s'
    formatter = logging.Formatter(FORMAT)
    hdlr.setFormatter(formatter)
    logger.addHandler(hdlr)
    logger.setLevel(logging.DEBUG)




# This bot takes random moves, chosen among the legal ones for its current
# position

TEAM_NAME = 'RandomBots'

def move(bot, state):
    # make a reference to our bot with a shorter name
    # mostly useful for longer code, of course
    return bot.random.choice(bot.legal_positions)



import pytest

from pathlib import Path
import py_compile
import sys
import tempfile

import pelita
from pelita.scripts.pelita_player import load_team_from_module, load_team

_mswindows = (sys.platform == "win32")

SIMPLE_MODULE = """
from pelita.player import stopping_player
TEAM_NAME = "%s"
move = stopping_player
"""

SIMPLE_FAILING_MODULE = """
def noteam():
    return None
"""

MODULE_IMPORT_ERROR = """
import .broken
TEAM_NAME = "noteam"
def move(bot, state):
    return bot.position
"""

# TODO: The modules should be unloaded after use
# If we import modules with the same name again, the results will be very unexpected

class TestLoadFactory:
    def test_simple_module_import(self):
        modules_before = list(sys.modules.keys())
        with tempfile.TemporaryDirectory() as d:
            module = Path(d) / "teamx"
            module.mkdir()
            initfile = module / "__init__.py"
            with initfile.open(mode='w') as f:
                f.write(SIMPLE_MODULE)

            spec = str(module)
            load_team_from_module(spec)

    def test_simple_file_import(self):
        modules_before = list(sys.modules.keys())
        with tempfile.TemporaryDirectory() as d:
            module = Path(d) / "teamy"
            module.mkdir()
            initfile = module / "teamyy.py"
            with initfile.open(mode='w') as f:
                f.write(SIMPLE_MODULE)

            spec = str(initfile)
            load_team_from_module(spec)

    def test_failing_import(self):
        modules_before = list(sys.modules.keys())
        with tempfile.TemporaryDirectory() as d:
            module = Path(d) / "teamz"
            module.mkdir()
            initfile = module / "__init__.py"
            with initfile.open(mode='w') as f:
                f.write(SIMPLE_FAILING_MODULE)

            spec = str(module)
            with pytest.raises(AttributeError):
                load_team_from_module(spec)

    def test_import_of_pyc(self):
        with tempfile.TemporaryDirectory() as d:
            module = Path(d) / "teampyc"
            module.mkdir()
            initfile = module / "teampycpyc.py"
            with initfile.open(mode='w') as f:
                f.write(SIMPLE_MODULE)
            pycfile = initfile.parent / "teampycpyc.pyc"
            py_compile.compile(str(initfile), cfile=str(pycfile))
            initfile.unlink()

            spec = str(pycfile)
            load_team_from_module(spec)

    def test_failing_import_importerror(self):
        with tempfile.TemporaryDirectory() as d:
            module = Path(d) / "teamzab"
            module.mkdir()
            initfile = module / "__init__.py"
            with initfile.open(mode='w') as f:
                f.write(MODULE_IMPORT_ERROR)
            broken_module = module / "broken.py"
            with broken_module.open(mode='w') as f:
                f.write('this is a syntax error\n')

            spec = str(module)

            with pytest.raises(SyntaxError):
                load_team_from_module(spec)

class TestLoadTeam:
    def test_simple_module_import_forbidden_names(self):
        names = ["", " ", "-", "∂", "0" * 26]
        for idx, name in enumerate(names):
            modules_before = list(sys.modules.keys())
            with tempfile.TemporaryDirectory() as d:
                module = Path(d) / ("teamx_%i" % idx)
                module.mkdir()
                initfile = module / "__init__.py"
                with initfile.open(mode='w') as f:
                    try:
                        f.write(SIMPLE_MODULE % (name,))
                    except UnicodeEncodeError:
                        if _mswindows:
                            # Ignore UnicodeEncodeErrors on Windows for this test
                            # It is too complicate to debug this
                            continue
                        else:
                            raise

                spec = str(module)
                with pytest.raises(ValueError):
                    load_team(spec)

    def test_simple_module_import_allowed_names(self):
        names = ["a", "a a", "0" * 25]
        for idx, name in enumerate(names):
            modules_before = list(sys.modules.keys())
            with tempfile.TemporaryDirectory() as d:
                module = Path(d) / ("teamy_%i" % idx)
                module.mkdir()
                initfile = module / "__init__.py"
                with initfile.open(mode='w') as f:
                    f.write(SIMPLE_MODULE % (name,))

                spec = str(module)
                load_team(spec)

    # These test cases need to be handled in one function
    # ie. not in a parametrized test, as the will need
    # to be run inside the same Python session
    load_team_cases = [
        ("pelita/player/StoppingPlayer", None),
#        ("StoppingPlayer,StoppingPlayer", None),
        ("NonExistingPlayer", ImportError),
#        ("StoppingPlayer,StoppingPlayer,FoodEatingPlayer", ValueError),
        #('doc/source/groupN', AttributeError), # TODO: Should be rewritten for a proper team
        #('doc/source/groupN/__init__.py', ImportError), # TODO: Should be rewritten for a proper team
        #('doc/source/groupN', ValueError), # Has already been imported
    ]

    def test_load_team(self):
        for path, result in self.load_team_cases:
            print(path, result)
            if result is not None:
                with pytest.raises(result):
                    load_team(path)
            else:
                load_team(path)





def smart_random_player(bot, state):
    dangerous_enemy_pos = [enemy.position
        for enemy in bot.enemy if enemy.position in enemy.homezone]
    killable_enemy_pos = [enemy.position
        for enemy in bot.enemy if enemy.position not in enemy.homezone]

    smart_positions = []
    for new_pos in bot.legal_positions[:]:
        if (new_pos == bot.position or
            new_pos in dangerous_enemy_pos):
            continue # bad idea
        elif (new_pos in killable_enemy_pos or
              new_pos in bot.enemy[0].food):
            return new_pos # get it
        else:
            smart_positions.append(new_pos)

    if smart_positions:
        return bot.random.choice(smart_positions)
    else:
        # we ran out of smart moves
        return bot.position


TEAM_NAME = "Smart Random Players"
move = smart_random_player



from random import random
import pytest

from pelita.game import run_game
from pelita.layout import parse_layout
from pelita.player import (SANE_PLAYERS, food_eating_player, nq_random_player,
                           random_explorer_player, smart_eating_player,
                           smart_random_player)
from pelita.utils import setup_test_game


class TestNQRandom_Player:
    def test_demo_players(self):
        test_layout = (
        """ ############
            #a#.   .# x#
            ###.   .####
            #b#.   .# y#
            ############ """)
        teams = [
            nq_random_player,
            nq_random_player
        ]
        state = run_game(teams, layout_dict=parse_layout(test_layout), max_rounds=1)
        assert state['bots'][0] == (1, 1)
        assert state['bots'][1] == (9, 1)

    def test_path(self):
        test_layout = (
        """ ############
            #  . # .# ##
            # ## #  # ##
            #a#.   .##x#
            ###.   .####
            #b#.   .# y#
            ############ """)
        teams = [
            nq_random_player,
            nq_random_player
        ]
        state = run_game(teams, layout_dict=parse_layout(test_layout), max_rounds=7)
        assert state['bots'][0] == (4, 3)
        assert state['bots'][1] == (10, 3)

    def test_only_move_forward(self):
        # test that the bot only moves forward
        layout="""
        ########
        #y #.###
        #b.x a #
        ########
        """
        bot = setup_test_game(layout=layout, is_blue=True)
        bot.track = [(6, 2), (5, 2)]
        assert bot.position == (5, 2)
        next_pos = nq_random_player(bot, {})
        assert next_pos == (4, 2)

    def test_only_move_unless_blocked(self):
        # test that the bot only moves forward
        layout="""
        ########
        #y #.###
        #b.x#a #
        ########
        """
        bot = setup_test_game(layout=layout, is_blue=True)
        bot.track = [(6, 2), (5, 2)]
        assert bot.position == (5, 2)
        next_pos = nq_random_player(bot, {})
        assert next_pos == (6, 2)


class TestRandomExplorerPlayers:
    def test_goes_to_unvisited(self):
        layout="""
        ########
        #y # .##
        #b.x a #
        ########
        """
        bot = setup_test_game(layout=layout, is_blue=True)
        state = {0: {'visited': [(5, 1), (4, 1), (4, 2), (3, 2), (2, 2), (1, 2)]}}
        assert bot.position == (5, 2)
        next_pos = random_explorer_player(bot, state)
        assert next_pos == (6, 2)

    def test_goes_to_least_recently_visited(self):
        layout="""
        ########
        #y # .##
        #b.x a #
        ########
        """
        bot = setup_test_game(layout=layout, is_blue=True)
        state = {0: {'visited': [(6, 2), (5, 2), (5, 1), (4, 1), (4, 2), (3, 2), (2, 2), (1, 2)]}}
        assert bot.position == (5, 2)
        # check that we have already visited all legal positions
        assert all(pos in state[0]['visited'] for pos in bot.legal_positions)
        next_pos = random_explorer_player(bot, state)
        assert next_pos == (4, 2)


class TestSmartRandomPlayer:
    def test_legalmoves(self):
        # check that the only two valid moves are always returned
        # we try ten times, to test 10 different random streams
        layout="""
        ########
        #a######
        #b. .xy#
        ########
        """
        for i in range(10):
            bot = setup_test_game(layout=layout, is_blue=True)
            next_pos = smart_random_player(bot, {})
            assert next_pos in ((1,2), (1,1))

    def test_kill_enemy(self):
        # check that we indeed kill an enemy when possible
        layout="""
        ########
        #x###.##
        #a.  by#
        ########
        """
        bot = setup_test_game(layout=layout, is_blue=True)
        next_pos = smart_random_player(bot, {})
        assert next_pos == (1,1)

    def test_eat_food(self):
        # check that we indeed collect food when possible
        layout="""
        ########
        #y # .##
        #b.x a #
        ########
        """
        bot = setup_test_game(layout=layout, is_blue=True)
        next_pos = smart_random_player(bot, {})
        assert next_pos == (5,1)

    def test_no_kamikaze_stop(self):
        # Check that we stop if escaping would kill us
        layout="""
        ########
        #  ###.#
        #b. xay#
        ########
        """
        bot = setup_test_game(layout=layout, is_blue=True)
        next_pos = smart_random_player(bot, {})
        assert next_pos == (5, 2)


class TestFoodEatingPlayer:
    def test_legalmoves(self):
        # check that the only two valid moves are returned
        layout="""
        ########
        #a######
        #b. .xy#
        ########
        """
        bot = setup_test_game(layout=layout, is_blue=True)
        next_pos = food_eating_player(bot, {})
        assert next_pos in ((1,2), (1,1))

    def test_eat_food(self):
        # check that we eat the last food when adjacent
        layout="""
        ########
        #x  a.##
        # .  by#
        ########
        """
        bot = setup_test_game(layout=layout, is_blue=True)
        next_pos = food_eating_player(bot, {})
        assert next_pos == (5,1)

    def test_move_towards_food(self):
        # check that we move closer to the food
        layout="""
        ########
        #y a .##
        #b.x   #
        ########
        """
        bot = setup_test_game(layout=layout, is_blue=True)
        next_pos = food_eating_player(bot, {})
        assert next_pos == (4,1)

    def test_move_towards_next_food(self):
        # check that we move closer to the food in the next_food key,
        # even though another food is closer
        layout="""
        ########
        #y   .##
        #      #
        # .. a #
        #x   . #
        # .b   #
        ########
        """
        bot = setup_test_game(layout=layout, is_blue=True)
        next_pos = food_eating_player(bot, {0: {'next_food': (5, 1)}})
        assert next_pos == (5, 2)

    def test_will_do_kamikaze(self):
        # check that we eat the last food when adjacent
        layout="""
        ########
        #x  a.##
        # .  b #
        ########
        """
        bot = setup_test_game(layout=layout, is_blue=True, bots={'y': (5, 1)})
        next_pos = food_eating_player(bot, {})
        assert next_pos == (5,1)

class TestSmartEatingPlayer:
    def test_legalmoves(self):
        # check that the only two valid moves are returned
        layout="""
        ########
        #a######
        #b. .xy#
        ########
        """
        bot = setup_test_game(layout=layout, is_blue=True)
        next_pos = smart_eating_player(bot, {})
        assert next_pos in ((1,2), (1,1))

    def test_eat_food(self):
        # check that we eat the last food when adjacent
        layout="""
        ########
        #x  a.##
        # .  by#
        ########
        """
        bot = setup_test_game(layout=layout, is_blue=True)
        next_pos = smart_eating_player(bot, {})
        assert next_pos == (5,1)

    def test_move_towards_food(self):
        # check that we move closer to the food
        layout="""
        ########
        #y a .##
        #b.x   #
        ########
        """
        bot = setup_test_game(layout=layout, is_blue=True)
        next_pos = smart_eating_player(bot, {})
        assert next_pos == (4,1)

    def test_move_towards_next_food(self):
        # check that we move closer to the food in the next_food key,
        # even though another food is closer
        layout="""
        ########
        #y   .##
        #      #
        # .. a #
        #x   . #
        # .b   #
        ########
        """
        bot = setup_test_game(layout=layout, is_blue=True)
        next_pos = smart_eating_player(bot, {0: {'next_food': (5, 1)}})
        assert next_pos == (5, 2)

    def test_will_not_do_kamikaze(self):
        # check that we eat the last food when adjacent
        layout="""
        ########
        #x  a.##
        # .  b #
        ########
        """
        bot = setup_test_game(layout=layout, is_blue=True, bots={'y': (5, 1)})
        next_pos = smart_eating_player(bot, {})
        assert next_pos == (4,1)

@pytest.mark.parametrize('player', SANE_PLAYERS)
def test_players(player):
    # Simple checks that the players are running to avoid API discrepancies
    test_layout = (
    """ ############
        #b . # . y##
        # ## #    ##
        # ## #  # ##
        # ## #  # ##
        #    #  # ##
        #a#.   .  x#
        ############ """)

    teams = [
        player,
        nq_random_player
    ]
    state = run_game(teams, layout_dict=parse_layout(test_layout), max_rounds=20)
    assert state['gameover']
    # ensure that all test players ran correctly
    assert state['fatal_errors'] == [[], []]
    # our test players should never return invalid moves
    assert state['errors'] == [{}, {}]




import random

import networkx as nx


from .team import make_bots, create_homezones
from .game import split_food, SHADOW_DISTANCE
from .layout import (get_random_layout, get_layout_by_name, get_available_layouts,
                     parse_layout, BOT_N2I, initial_positions)
from .gamestate_filters import manhattan_dist

# this import is needed for backward compatibility, do not remove or you'll break
# older clients!
from .team import walls_to_graph

RNG = random.Random()


def _parse_layout_arg(*, layout=None, food=None, bots=None, seed=None):

    # prepare layout argument to be passed to pelita.game.run_game
    if layout is None:
        layout_name, layout_str = get_random_layout(size='normal', seed=seed)
        layout_dict = parse_layout(layout_str)
    elif layout in get_available_layouts(size='all'):
        # check if this is a built-in layout
        layout_name = layout
        layout_str = get_layout_by_name(layout)
        layout_dict = parse_layout(layout_str)
    else:
        # OK, then it is a (user-provided) layout string
        layout_str = layout
        layout_name = '<string>'
        # be strict and complain if the layout does not contain two bots and two enemies
        layout_dict = parse_layout(layout_str, food=food, bots=bots)

    return layout_dict, layout_name


# this is a dumbed-down version of pelita.game.run_game, useful to be exposed to the
# users to run background games. It hides most of the parameters of run_game which
# are not relevant to the user in this setup and reformats the rest so that we
# don't leak internal implementation details and indices.
def run_background_game(*, blue_move, red_move, layout=None, max_rounds=300, seed=None):
    """Run a pelita match.

    Parameters
    ----------
    blue_move : function
             the move function that will be used by the blue team

    red_move : function
             the move function that will be used by the red team

    layout : str
          specify the layout of the maze to play with. If None, a built-in
          layout of normal size will be chosen at random. If specified it will
          be interpreted as the name of a built-in layout, e.g. 'normal_083'.
          You can also pass a layout string as in:
          '''
          ########
          #. b y #
          #a x   #
          ########
          '''

    max_rounds : int
              maximum number of rounds to play before the game is over. Default: 300.

    seed : int
        seed used to initialize the random number generator.


    Returns
    -------
    game_state : dict
              the final game state as a dictionary. Dictionary keys are:
              - 'seed' : the seed used to initialize the random number generator
              - 'walls' : set of wall coordinates for the layout
              - 'layout' : the name of the used layout
              - 'round' : the round at which the game was over
              - 'draw' : True if the game ended in a draw
              - 'blue_food' : list of food coordinates for the blue team. Note that
                              this can be empty if the red team has eaten all the
                              blue team's food
             - 'red_food' : list of food coordinates for the red team
             - 'blue_bots' : list of coordinates for the blue team's bots
             - 'red_bots' : list of coordinates for the red team's bots
             - 'blue_score' : score of the blue team
             - 'red_score' : score of the red team
             - 'blue_errors' : a dictionary collecting non-fatal errors for the
                               blue team
             - 'red_errors' : a dictionary collecting non-fatal errors for the
                              red team
             - 'blue_deaths' : a list with the number of times each bot in the blue
                               team has been killed
             - 'red_deaths' : a list with the number of times each bot in the red
                              team has been killed
             - 'blue_kills' : a list with the number of times each bot in the blue
                              team has killed enemy bots
             - 'red_kills' : a list with the number of times each bot in the red
                              team has killed enemy bots
             - 'blue_wins' : True if the blue team wins
             - 'red_wins' : True if the red team wins

    Notes
    -----
    - Move functions:
        A move function is a function with signature move(bot, state) -> (x, y).
        The function takes a bot object and a state and returns the next position of
        the current bot as a tuple (x, y). `state` is initially an empty dictionary.

    - Timeouts:
        As opposed to standard pelita matches, timeouts are not considered.

    """
    from .game import run_game

    # if the seed is not set explicitly, set it here
    if seed is None:
        seed = RNG.randint(1, 2**31)
        RNG.seed(seed)

    layout_dict, layout_name = _parse_layout_arg(layout=layout, seed=seed)

    game_state = run_game((blue_move, red_move), layout_dict=layout_dict,
                          layout_name=layout_name, max_rounds=max_rounds, seed=seed,
                          team_names=('blue', 'red'), allow_exceptions=True, print_result=False)
    out = {}
    out['seed'] = seed
    out['walls'] = game_state['walls']
    out['round'] = game_state['round']
    out['layout'] = layout_name
    out['blue_food'] = list(game_state['food'][0])
    out['red_food'] = list(game_state['food'][1])
    out['blue_bots'] = game_state['bots'][::2]
    out['red_bots'] = game_state['bots'][1::2]
    out['blue_score'] = game_state['score'][0]
    out['red_score'] = game_state['score'][1]
    out['blue_errors'] = game_state['errors'][0]
    out['red_errors'] = game_state['errors'][1]
    out['blue_deaths'] = game_state['deaths'][::2]
    out['red_deaths'] = game_state['deaths'][1::2]
    out['blue_kills'] = game_state['kills'][::2]
    out['red_kills'] = game_state['kills'][1::2]
    out['blue_wins'], out['red_wins'], out['draw'] = False, False, False
    if game_state['whowins'] == 0:
        out['blue_wins'] = True
    elif game_state['whowins'] == 1:
        out['red_wins'] = True
    else:
        out['draw'] = True

    return out


def shaded_food(pos, food, radius):
    # Get all food that is in a radius around any of pos
    # TODO: This duplicates code in update_food_age
    for pellet in food:
        if any(manhattan_dist(ghost, pellet) <= radius for ghost in pos):
            yield pellet


def setup_test_game(*, layout, is_blue=True, round=None, score=None, seed=None,
                    food=None, bots=None, is_noisy=None):
    """Setup a test game environment useful for testing move functions.

    Parameters
    ----------
    layout : str
      specify the layout of the maze to play with. If None, a built-in
      layout of normal size will be chosen at random. If specified it will
      be interpreted as the name of a built-in layout, e.g. 'normal_083'.

    is_blue : bool
           when True, sets up up the game assuming your bots are in the blue team,
           i.e. the team with the homezone on the left side of the maze. When False,
           your bots are assumed to be in the red team.

    round : int
         set the current round for the returned game state

    score : list[int, int]
         set the score for you and the enemy team. Example: [12,15] -> your score
         is 12, your enemy score is 15. If None, score will be set to [0, 0]

    seed : int
        set the random number generator seed to get reproducible results if your
        move function uses the bot.random random generator

    food : list[(x0,y0), (x1,y1),...]
        list of coordinates for food pellets. The food will be added to the one
        already found in the layout string

    bots : dict{"a": (a_x,a_y), "b":(b_x,b_y), "x":(x_x,x_y), "y":(y_x,y_y)}
           dict of bot names and coordinates. The items found here  will override
           the positions found in the layout.

    is_noisy : dict{"a": True, "b": False, "x": True, "y": False}
              Dict of bot names and booleans for the bots' is_noisy property.


    Returns
    -------
    bot : Bot
       a Bot object suitable to be passed to a move function
    """

    if score is None:
        score = [0, 0]

    # grab is_noisy overrides from user
    is_noisy_default = {char:False for char in BOT_N2I}
    if is_noisy is not None:
        is_noisy_default.update(is_noisy)

    is_noisy = is_noisy_default


    layout, layout_name = _parse_layout_arg(layout=layout, food=food, bots=bots)

    width, height = layout['shape']

    food = split_food(width, layout['food'])

    if is_blue:
        team_index = 0
        enemy_index = 1
        bot_positions = [layout['bots'][0], layout['bots'][2]]
        enemy_positions = [layout['bots'][1], layout['bots'][3]]
        is_noisy_enemy = [is_noisy["x"], is_noisy["y"]]
    else:
        team_index = 1
        enemy_index = 0
        bot_positions = [layout['bots'][1], layout['bots'][3]]
        enemy_positions = [layout['bots'][0], layout['bots'][2]]
        is_noisy_enemy = [is_noisy["a"], is_noisy["b"]]

    rng = random.Random(seed)

    team = {
        'bot_positions': bot_positions,
        'team_index': team_index,
        'score': score[team_index],
        'kills': [0]*2,
        'deaths': [0]*2,
        'bot_was_killed' : [False]*2,
        'error_count': 0,
        'food': food[team_index],
        'shaded_food': list(shaded_food(bot_positions, food[team_index], radius=SHADOW_DISTANCE)),
        'name': "blue" if is_blue else "red",
        'team_time': 0.0,
    }
    enemy = {
        'bot_positions': enemy_positions,
        'team_index': enemy_index,
        'score': score[enemy_index],
        'kills': [0]*2,
        'deaths': [0]*2,
        'bot_was_killed': [False]*2,
        'error_count': 0,
        'food': food[enemy_index],
        'shaded_food': [],
        'is_noisy': is_noisy_enemy,
        'name': "red" if is_blue else "blue",
        'team_time': 0.0,
    }

    bot = make_bots(walls=layout['walls'],
                    shape=layout['shape'],
                    initial_positions=initial_positions(layout['walls'], layout['shape']),
                    homezone=create_homezones(layout['shape'], layout['walls']),
                    team=team,
                    enemy=enemy,
                    round=round,
                    bot_turn=0,
                    rng=rng,
                    graph=walls_to_graph(layout['walls']))
    return bot




import pytest

from pelita import utils
from pelita.player import stopping_player


@pytest.mark.parametrize('is_blue', [True, False])
def test_setup_test_game(is_blue):
    # Test that is_noisy is set properly
    layout = """
    ##################
    #. ... .##.     y#
    # # #  .  .### # #
    # # ##. x .      #
    #      .   .## # #
    #a# ###.  .  # # #
    #b     .##. ... .#
    ##################
    """
    test_game = utils.setup_test_game(layout=layout, is_blue=is_blue, is_noisy={"a":False, "b":True, "x":False, "y":True})

    if is_blue:
        assert test_game.position == (1, 5)
        assert test_game.other.position == (1, 6)
        assert test_game.enemy[0].position == (8, 3)
        assert test_game.enemy[1].position == (16, 1)
    else:
        assert test_game.position == (8, 3)
        assert test_game.other.position == (16, 1)
        assert test_game.enemy[0].position == (1, 5)
        assert test_game.enemy[1].position == (1, 6)

    # load_builtin_layout loads unnoised enemies
    assert test_game.enemy[0].is_noisy is False
    assert test_game.enemy[1].is_noisy is True


@pytest.mark.parametrize('is_blue', [True, False])
def test_setup_test_game_incomplete_noisy_dict(is_blue):
    # Test that is_noisy is set properly
    layout = """
    ##################
    #. ... .##.     y#
    # # #  .  .### # #
    # # ##. x .      #
    #      .   .## # #
    #a# ###.  .  # # #
    #b     .##. ... .#
    ##################
    """
    test_game = utils.setup_test_game(layout=layout, is_blue=is_blue, is_noisy={"b":True, "y":True})
    # load_builtin_layout loads unnoised enemies
    assert test_game.enemy[0].is_noisy is False
    assert test_game.enemy[1].is_noisy is True


def test_run_background_game():
    result = utils.run_background_game(blue_move=stopping_player, red_move=stopping_player)
    result.pop('seed')
    result.pop('walls')
    result.pop('layout')
    result.pop('blue_food')
    result.pop('red_food')
    assert result == {
        'round': 300,
        'blue_bots': [(1, 13), (1, 14)],
        'red_bots': [(30, 2), (30, 1)],
        'blue_score': 0,
        'red_score': 0,
        'blue_errors': {},
        'red_errors': {},
        'blue_deaths': [0, 0],
        'red_deaths': [0, 0],
        'blue_kills': [0, 0],
        'red_kills': [0, 0],
        'blue_wins': False,
        'red_wins': False,
        'draw': True
    }




import pytest

import ci_engine


@pytest.fixture
def db_wrapper():
    wrapper = ci_engine.DB_Wrapper(':memory:')
    wrapper.create_tables()
    return wrapper

"""Tests for the DB_Wrapper class."""

def test_foreign_keys_enabled(db_wrapper):
    result = db_wrapper.cursor.execute("PRAGMA foreign_keys;").fetchone()
    assert result[0] == 1

def test_add_player(db_wrapper):
    db_wrapper.add_player('p1', 'h1')
    db_wrapper.add_player('p2', 'h2')
    with pytest.raises(ValueError):
        db_wrapper.add_player('p1', 'h1')
    players = sorted(db_wrapper.get_players())
    assert players == ['p1', 'p2']

def test_remove_player(db_wrapper):
    db_wrapper.add_player('p1', 'h1')
    db_wrapper.add_player('p2', 'h2')
    db_wrapper.add_player('p3', 'h3')
    db_wrapper.add_gameresult('p1', 'p2', 0, '{}', '', '')
    db_wrapper.add_gameresult('p2', 'p1', 0, '{}', '', '')
    db_wrapper.add_gameresult('p2', 'p3', 0, '{}', '', '')
    # player2 has three games
    assert len(db_wrapper.get_results('p2')) == 3
    db_wrapper.remove_player('p1')
    # player 1 should have no game results
    assert db_wrapper.get_results('p1') == []
    # after removing all games of player one, player2 should have 1
    # game
    assert len(db_wrapper.get_results('p2')) == 1
    # player 3 should be untouched
    assert len(db_wrapper.get_results('p3')) == 1

def test_add_remove_weirdly_named_player(db_wrapper):
    stupid_names = [
        "Little'",
        'Bobby"',
        "таблицы",
    ]

    for name in stupid_names:
        db_wrapper.add_player(name, name)
        db_wrapper.remove_player(name)

def test_get_players(db_wrapper):
    players = ['p1', 'p2', 'p3']
    for p in players:
        db_wrapper.add_player(p, 'h')
    players2 = sorted(db_wrapper.get_players())
    assert players == players2

def test_get_results(db_wrapper):
    db_wrapper.add_player('p1', 'h1')
    db_wrapper.add_player('p2', 'h2')
    # empty list if no results are available
    assert db_wrapper.get_results('p1') == []
    db_wrapper.add_gameresult('p1', 'p2', 0, '{}', '', '')
    result = db_wrapper.get_results('p1')[0]
    # check for correct values
    assert result[0] == 'p1'
    assert result[1] == 'p2'
    assert result[2] == 0
    db_wrapper.add_gameresult('p2', 'p1', 0, '{}', '', '')
    # check for correct number of results
    results = db_wrapper.get_results('p1')
    assert len(results) == 2

def test_get_player_hash(db_wrapper):
    db_wrapper.add_player('p1', 'h1')
    db_wrapper.add_player('p2', 'h2')
    with pytest.raises(ValueError):
        db_wrapper.get_player_hash('p0')
    assert db_wrapper.get_player_hash('p1') == 'h1'
    assert db_wrapper.get_player_hash('p2') == 'h2'

def test_get_team_name(db_wrapper):
    db_wrapper.add_player('p1', 'h1')
    db_wrapper.add_player('p2', 'h2')
    db_wrapper.add_team_name('p1', 'tn1')
    db_wrapper.add_team_name('p2', 'tn2')
    db_wrapper.add_team_name('p2', 'tn3') # check that override works
    with pytest.raises(ValueError):
        db_wrapper.get_team_name('p0')
    assert db_wrapper.get_team_name('p1') == 'tn1'
    assert db_wrapper.get_team_name('p2') == 'tn3'

def test_wins_losses(db_wrapper):
    db_wrapper.add_player('p1', 'h1')
    db_wrapper.add_player('p2', 'h2')
    db_wrapper.add_player('p3', 'h3')
    db_wrapper.add_gameresult('p1', 'p2', 0, "{}", "", "")
    assert db_wrapper.get_wins_losses() == [
        ('p1', 'p2', 1, 0, 0),
        ('p2', 'p1', 0, 1, 0)
    ]
    db_wrapper.add_gameresult('p1', 'p2', -1, "{}", "", "")
    assert db_wrapper.get_wins_losses() == [
        ('p1', 'p2', 1, 0, 1),
        ('p2', 'p1', 0, 1, 1)
    ]
    db_wrapper.add_gameresult('p2', 'p1', 1, "{}", "", "")
    assert db_wrapper.get_wins_losses() == [
        ('p1', 'p2', 2, 0, 1),
        ('p2', 'p1', 0, 2, 1)
    ]
    db_wrapper.add_gameresult('p3', 'p1', 1, "{}", "", "")
    assert db_wrapper.get_wins_losses() == [
        ('p1', 'p2', 2, 0, 1),
        ('p1', 'p3', 1, 0, 0),
        ('p2', 'p1', 0, 2, 1),
        ('p3', 'p1', 0, 1, 0)
    ]

    assert db_wrapper.get_wins_losses('p1') == [
        ('p1', 'p2', 2, 0, 1),
        ('p1', 'p3', 1, 0, 0),
    ]

    assert db_wrapper.get_results('p1') == [
        ('p1', 'p2', 0),
        ('p1', 'p2', -1),
        ('p2', 'p1', 1),
        ('p3', 'p1', 1)
    ]

    assert db_wrapper.get_game_count('p1') == 4
    assert db_wrapper.get_game_count('p2') == 3
    assert db_wrapper.get_game_count('p1', 'p2') == 3
    assert db_wrapper.get_game_count('p2', 'p1') == 3
    assert db_wrapper.get_game_count('p3', 'p1') == 1



#!/usr/bin/env python3

import argparse
import json
import logging
from pathlib import Path
import random
import sys
from urllib.parse import urlparse

from rich.console import Console
from rich.markup import escape
from rich.prompt import Prompt
import zmq

import pelita
from .script_utils import start_logging

from pelita.network import PELITA_PORT
# TODO: The check_team option
from pelita.tournament import check_team

# silence stupid warnings from logging module
logging.root.manager.emittedNoHandlerWarning = 1
_logger = logging.getLogger(__name__)


def scan(team_spec):
    if team_spec != "SCAN":
        return team_spec
    from zeroconf import ServiceBrowser, ServiceStateChange, Zeroconf
    from queue import Queue, Empty

    SCAN_TIME = 5 # seconds

    q = Queue(maxsize=20)

    def on_service_state_change(
        zeroconf: Zeroconf, service_type: str, name: str, state_change: ServiceStateChange
    ) -> None:

        if state_change is ServiceStateChange.Added:
            info = zeroconf.get_service_info(service_type, name)

            def make_url(addr, port):
                if port == PELITA_PORT:
                    return f"pelita://{addr}"
                else:
                    return f"pelita://{addr}:{port}"

            if info:
                addresses = [make_url(addr, info.port) for addr in info.parsed_scoped_addresses()]

                team_name = info.properties[b"team_name"].decode()
                path = info.properties[b"path"].decode()
                q.put((addresses[0] + path, team_name), timeout=5)

    zeroconf = Zeroconf()
    services = ["_pelita-player._tcp.local."]
    console = Console()

    console.print(f"[bold]Remote player requested. Scanning network for players ({SCAN_TIME}s).")
    with console.status("[red bold]Searching for other players …") as status:
        try:
            browser = ServiceBrowser(zeroconf, services, handlers=[on_service_state_change])
            players = []
            import time
            import select
            import sys
            start = time.time()
            while start + 5 > time.time():
                time.sleep(0.2)
                (addr, team_name) = q.get(timeout=5)
                #  if not players:
                    #console.print("[bold]Found players:")
                console.print(f"  [blue]{len(players)})[/] {team_name} \\[[blue]{addr}[/]]", highlight=False)
                players.append(f"remote:{addr}")
        except Empty:
            pass
        except KeyboardInterrupt:
            return None
        finally:
            zeroconf.close()
    if players:
        console.print(f"  [blue]r)[/] Random team")
        console.print(f"  [blue]x)[/] Exit")
        console.print()
        console.print(f"Found {len(players)} player{'s' if len(players) != 1 else ''}.")

        choices = {str(i): player for i, player in enumerate(players)}

        answer = Prompt.ask("[bold]Select player to play against (r for random, x to exit)",
                            choices=list(choices) + ["r", "x"],
                            default="r")

        if answer == "r":
            console.print("Choosing random player.")
            return random.choice(players)
        elif answer in choices.keys():
            console.print(f"Choosing [blue]{choices[answer]}[/]", highlight=False)
            return choices[answer]
        else:
            return None


def scan_server(team_spec):
    parsed_url = urlparse(team_spec)
    if parsed_url.port:
        port = parsed_url.port
    else:
        port = PELITA_PORT

    send_addr = f"tcp://{parsed_url.hostname}:{port}"

    zmq_context = zmq.Context()
    socket = zmq_context.socket(zmq.DEALER)
    socket.setsockopt(zmq.LINGER, 0)
    socket.connect(send_addr)
    socket.send_json({"SCAN": team_spec})

    console = Console()

    console.print(f"[bold]Remote player requested. Scanning server for players.")

    WAIT_TIMEOUT = 5000
    incoming = socket.poll(timeout=WAIT_TIMEOUT)
    if incoming == zmq.POLLIN:
        teams = json.loads(socket.recv().decode('utf8'))
        if not teams:
            console.print("No teams found on the server :(")
            return None
        else:
            print("Server has the following teams available")
    else:
        console.print(f"Server did not reply in {WAIT_TIMEOUT} ms.")
        return None

    players = []
    for addr, team_name in teams.items():
        console.print(f"  [blue]{len(players)})[/] {team_name} \\[[blue]{addr}[/]]", highlight=False)
        players.append(addr)

    if players:
        console.print(f"  [blue]r)[/] Random team")
        console.print(f"  [blue]s)[/] Server default")
        console.print(f"  [blue]x)[/] Exit")
        console.print()
        console.print(f"Found {len(players)} player{'s' if len(players) != 1 else ''}.")

        choices = {str(i): player for i, player in enumerate(players)}

        answer = Prompt.ask("[bold]Select player to play against (r for random, s server’s choice, x to exit)",
                            choices=list(choices) + ["r", "s", "x"],
                            default="r")

        if answer == "r":
            console.print("Choosing random player.")
            return random.choice(players)
        if answer == "s":
            console.print("Letting the server choose.")
            return team_spec
        elif answer in choices.keys():
            console.print(f"Choosing [blue]{choices[answer]}[/]", highlight=False)
            return choices[answer]
        else:
            return None


def geometry_string(s):
    """Get a X-style geometry definition and return a tuple.

    600x400 -> (600,400)
    """
    try:
        x_string, y_string = s.split('x')
        geometry = (int(x_string), int(y_string))
    except ValueError:
        msg = "%s is not a valid geometry specification" %s
        raise argparse.ArgumentTypeError(msg) from None
    return geometry


def long_help(s):
    return s if '--long-help' in sys.argv else argparse.SUPPRESS

parser = argparse.ArgumentParser(description='Run a single pelita game',
                                 add_help=False,
                                 formatter_class=argparse.RawDescriptionHelpFormatter)
parser._positionals = parser.add_argument_group('Arguments')
parser.add_argument('team_specs', help='FILENAME1.py FILENAME2.py (see below)', nargs='*', default=None)

parser._optionals = parser.add_argument_group('Options')
help_opt = parser._optionals.add_mutually_exclusive_group()
help_opt.add_argument('--help', '-h', help='Show this help message and exit.',
                    action='store_const', const=True)
help_opt.add_argument('--long-help', help='Show all possible options and exit.',
                    action='store_const', const=True)

parser.add_argument('--version', help='Show the version number and exit.',
                    action='store_const', const=True)
parser.add_argument('--log', help='Print debugging log information to'
                    ' LOGFILE (default \'stderr\').',
                    metavar='LOGFILE', const='-', nargs='?')
parser.add_argument('--write-replay', help=long_help('Print game dumps to file (will be overwritten)'),
                    metavar='REPLAYFILE', const='pelita.dump', nargs='?')
parser.add_argument('--replay', help=long_help('Replay a dumped game'),
                    metavar='REPLAYFILE', dest='replayfile', const='pelita.dump', nargs='?')
parser.add_argument('--store-output', help=long_help('Write all player’s stdout/stderr to the given folder (must exist)'),
                    metavar='FOLDER')
parser.add_argument('--list-layouts', action='store_true',
                    help='List all available built-in layouts.')
parser.add_argument('--check-team', action="store_true",
                    help=long_help('Check that the team is valid (on first sight) and print its name.'))
parser.add_argument('--append-blue', type=str, metavar='INFO', default=None,
                    help=long_help('Append info about the blue team (such as group id).'))
parser.add_argument('--append-red', type=str, metavar='INFO', default=None,
                    help=long_help('Append info about the red team (such as group id).'))

game_settings = parser.add_argument_group('Game settings')
game_settings.add_argument('--rounds', type=int, default=300,
                           help='Maximum number of rounds to play.')
game_settings.add_argument('--seed', type=int, metavar='SEED', default=None,
                           help='Initialize the random number generator with SEED.')
game_settings.add_argument('--allow-camping', const=True, action='store_const',
                           help='Food does not age when in a bot’s shadow')

layout_opt = game_settings.add_mutually_exclusive_group()
layout_opt.add_argument('--layout', metavar='LAYOUT',
                        help='Use maze layout specified in LAYOUT. LAYOUT can be'
                             ' a file containing a valid layout or the name of a '
                             'built-in layout. Use --list-layouts to get a list '
                             'of all available layouts.')
layout_opt.add_argument('--size', metavar='STRING', default='normal',
                        help="Pick a random maze layout of specified size."
                        " Possible sizes: 'small' (16x8), 'normal' (32x16), 'big' (64x32), 'all' (any of the previous). Default: 'normal'")

timeout_opt = game_settings.add_mutually_exclusive_group()
timeout_opt.add_argument('--timeout', type=float, metavar="SEC",
                         dest='timeout_length', help='Time before timeout is triggered (default: 3 seconds).')
timeout_opt.add_argument('--no-timeout', const=None, action='store_const',
                         dest='timeout_length', help='Run game without timeouts.')
game_settings.add_argument('--error-limit', type=int, default=5,
                           dest='error_limit', help='Error limit. Reaching this limit disqualifies a team (default: 5).')
parser.set_defaults(timeout_length=3)
game_settings.add_argument('--stop-at', dest='stop_at', type=int, metavar="N",
                           help='Stop before playing round N.')

viewer_settings = parser.add_argument_group('Viewer settings')
viewer_settings.add_argument('--geometry', type=geometry_string, metavar='NxM',
                    help='Set initial size of the game window.')
viewer_settings.add_argument('--fullscreen', const=True, action='store_const',
                    help='Make the game window run fullscreen')
viewer_settings.add_argument('--fps', type=float, default=40,
                    help='Set (approximate) number of frames per second in a graphical viewer.')

viewer_opt = viewer_settings.add_mutually_exclusive_group()
viewer_opt.add_argument('--null', action='store_const', const='null',
                        dest='viewer', help='Use no viewer on stdout.')
viewer_opt.add_argument('--ascii', action='store_const', const='ascii',
                        dest='viewer', help=long_help('Use the ASCII viewer.'))
viewer_opt.add_argument('--progress', action='store_const', const='progress',
                        dest='viewer', help=long_help('Use the progress viewer.'))
viewer_opt.add_argument('--tk', action='store_const', const='tk',
                        dest='viewer', help='Use the tk viewer (default).')
viewer_opt.add_argument('--tk-no-sync', action='store_const', const='tk-no-sync',
                        dest='viewer', help=long_help('Uses the tk viewer in an unsynchronized mode.'))
parser.set_defaults(viewer='tk')

advanced_settings = parser.add_argument_group('Advanced settings')
advanced_settings.add_argument('--reply-to', type=str, metavar='URL', dest='reply_to',
                               help=long_help('Communicate the result of the game on this channel.'))

publisher_opt = advanced_settings.add_mutually_exclusive_group()
publisher_opt.add_argument('--publish', type=str, metavar='URL', dest='publish_to',
                           help=long_help('Publish the game to this zmq socket.'))
publisher_opt.add_argument('--no-publish', const=False, action='store_const', dest='publish_to',
                           help=long_help('Do not publish.'))
parser.set_defaults(publish_to="tcp://127.0.0.1")

advanced_settings.add_argument('--controller', type=str, metavar='URL', default="tcp://127.0.0.1",
                               help=long_help('Channel for controlling the game.'))
advanced_settings.add_argument('--external-controller', const=True, action='store_const',
                               help=long_help('Force control by an external controller.'))

parser.epilog = """\
Team Specification:
    A team consists of a path to a .py file or to a Python module
    that defines at least:

    * TEAM_NAME
        a string with the name of the team.

    * move(bot, state) -> next_position
        a function which given a bot and a state returns the next position for
        current bot and a state.

    Example file: my_stopping_bots.py

        TEAM_NAME = 'My stopping bots'

        def move(bot, state):
            return bot.position

    A game between two teams of stopping bots can then be played as

        pelita my_stopping_bots.py my_stopping_bots.py

    Demo players can be found at https://github.com/ASPP/pelita_template

Layout specification:
    If --layout is not specified, the maze is chosen at random from the pool of
    built-in normal-sized layouts. You can change this pool by using --size.
"""


def main():
    args = parser.parse_args()
    if args.help or args.long_help:
        parser.print_help()
        sys.exit(0)

    if args.version:
        print("Pelita {}".format(pelita.__version__))
        sys.exit(0)

    if args.list_layouts:
        layouts = pelita.layout.get_available_layouts(size='all')
        layouts.sort()
        print('\n'.join(layouts))
        sys.exit(0)

    if args.viewer.startswith('tk') and not args.publish_to:
        raise ValueError("Options --tk (or --tk-no-sync) and --no-publish are mutually exclusive.")

    if args.log:
        start_logging(args.log)

    if args.rounds < 1:
        raise ValueError(f"Must play at least one round (rounds={args.rounds}).")

    if args.check_team:
        if not args.team_specs:
            raise ValueError("No teams specified.")
        for team_spec in args.team_specs:
            try:
                team_name = check_team(team_spec)
                print("NAME:", team_name)
            except pelita.network.ZMQClientError as e:
                if e.error_type == 'ModuleNotFoundError':
                    #print(f"{e.message}")
                    pass
                else:
                    raise
        sys.exit(0)

    if args.viewer == 'null':
        viewers = []
    else:
        viewers = [args.viewer]

    geometry = args.geometry
    delay = int(1000./args.fps)
    stop_at = args.stop_at

    viewer_options = {
        "fullscreen" : args.fullscreen,
        "geometry": geometry,
        "delay": delay,
        "stop_at": stop_at
    }

    if args.reply_to:
        viewers.append(('reply-to', args.reply_to))
    if args.write_replay:
        viewers.append(('write-replay-to', args.write_replay))

    if args.replayfile:
        viewer_state = pelita.game.setup_viewers(viewers, options=viewer_options)
        if pelita.game.controller_exit(viewer_state, await_action='set_initial'):
            sys.exit(0)

        old_game = Path(args.replayfile).read_text().split("\x04")
        for state in old_game:
            if not state.strip():
                continue
            state = json.loads(state)
            # walls, bots, food must be list of tuple
            state['walls'] = list(map(tuple, state['walls']))
            state['bots'] = list(map(tuple, state['bots']))
            state['food'] = list(map(tuple, state['food']))
            for viewer in viewer_state['viewers']:
                viewer.show_state(state)
            if pelita.game.controller_exit(viewer_state):
                break

        sys.exit(0)

    # Run a normal game
    num_teams = 2
    team_specs = args.team_specs
    if len(team_specs) == 0:
        team_specs = ('0', '1')
    if len(team_specs) == 1:
        raise RuntimeError("Not enough teams given. Must be {}".format(num_teams))
    if len(team_specs) > num_teams:
        raise RuntimeError("Too many teams given. Must be < {}.".format(num_teams))

    for idx, team_spec in enumerate(team_specs):
        if team_spec == "SCAN":
            scanned_spec = scan(team_spec)
            if scanned_spec:
                team_specs[idx] = scanned_spec
            else:
                print("No remote team found. Exiting.")
                return
        elif team_spec.startswith("pelita://"):
            # check if we need to send a server scan request
            parsed_url = urlparse(team_spec)
            if parsed_url.path in ('', '/'):
                scanned_spec = scan_server(team_spec)
                if scanned_spec:
                    team_specs[idx] = scanned_spec
                else:
                    print("No remote team selected. Exiting.")
                    return

    if args.seed is None:
        seed = random.randint(0, sys.maxsize)
        print("Replay this game with --seed {seed}".format(seed=seed))
    else:
        seed = args.seed

    random.seed(seed)

    if args.layout:
        # first check if the given layout is a file
        layout_path = Path(args.layout)
        if layout_path.exists():
            # OK, this is a valid file, load it
            layout_name = str(layout_path)
            # use the basename of the path as a layout name
            layout_name = layout_path.parts[-1]
            layout_string = layout_path.read_text()
        else:
            # if the given layout is not a path, we assume it is the name of one
            # of the built-in paths
            layout_name = args.layout
            layout_string = pelita.layout.get_layout_by_name(args.layout)
    else:
        layout_name, layout_string = pelita.layout.get_random_layout(args.size, seed=seed)

    print("Using layout '%s'" % layout_name)

    layout_dict = pelita.layout.parse_layout(layout_string)
    pelita.game.run_game(team_specs=team_specs, max_rounds=args.rounds, layout_dict=layout_dict, layout_name=layout_name, seed=seed,
                         allow_camping=args.allow_camping, timeout_length=args.timeout_length, error_limit=args.error_limit,
                         viewers=viewers, viewer_options=viewer_options,
                         store_output=args.store_output,
                         team_infos=(args.append_blue, args.append_red))

if __name__ == '__main__':
    main()



import pytest

import uuid
import sys

import zmq

from pelita.network import bind_socket
from pelita.team import make_team
from pelita.scripts.pelita_player import player_handle_request

_mswindows = (sys.platform == "win32")


@pytest.fixture(scope="module")
def zmq_context():
    context = zmq.Context()
    yield context
    context.destroy()

@pytest.mark.skipif(_mswindows, reason="No IPC sockets on Windows.")
def test_bind_socket_success(zmq_context):
    socket = zmq_context.socket(zmq.PUB)
    address = "ipc:///tmp/pelita-test-bind-socket-%s" % uuid.uuid4()
    bind_socket(socket, address)

def test_bind_socket_fail(zmq_context):
    socket = zmq_context.socket(zmq.PUB)
    with pytest.raises(zmq.ZMQError):
        bind_socket(socket, "bad-address", '--publish')
    socket.close()


def test_simpleclient(zmq_context):
    res = []
    def stopping(bot, state):
        print(bot)
        res.append(True)
        print(res)
        return bot.position

    sock = zmq_context.socket(zmq.PAIR)
    port = sock.bind_to_random_port('tcp://127.0.0.1')

    team, _ = make_team(stopping, team_name='test stopping player')
    client_sock = zmq_context.socket(zmq.PAIR)
    client_sock.connect(f'tcp://127.0.0.1:{port}')

    # Faking some data
    _uuid = uuid.uuid4().__str__()
    set_initial = {
        '__uuid__': _uuid,
        '__action__': "set_initial",
        '__data__': {
            'team_id': 0,
            'game_state': {
                'seed': 0,
                'walls': [(0, 0), (1, 0), (2, 0), (3, 0),
                          (0, 1),                 (3, 1),
                          (0, 2),                 (3, 2),
                          (0, 3), (1, 3), (2, 3), (3, 3),
                          ],
                'shape': (4, 4)
            }
        }
    }
    sock.send_json(set_initial)
    player_handle_request(client_sock, team)

    assert sock.recv_json() == {
        '__uuid__': _uuid,
        '__return__': 'test stopping player'
    }

    _uuid = uuid.uuid4().__str__()
    get_move = {
        '__uuid__': _uuid,
        '__action__': "get_move",
        '__data__': {
            'game_state': {
                'team': {
                    'team_index': 0,
                    'bot_positions': [(1, 1), (1, 1)],
                    'score': 0,
                    'kills': [0]*2,
                    'deaths': [0]*2,
                    'bot_was_killed': [False]*2,
                    'error_count': 0,
                    'food': [(1, 1)],
                    'shaded_food': [(1, 1)],
                    'name': 'dummy',
                    'team_time': 0,
                },
                'enemy': {
                    'team_index': 1,
                    'bot_positions': [(2, 2), (2, 2)],
                    'score': 0,
                    'kills': [0]*2,
                    'deaths': [0]*2,
                    'bot_was_killed': [False]*2,
                    'food': [(2, 2)],
                    'shaded_food': [],
                    'name': 'other dummy',
                    'team_time': 0,
                    'is_noisy': [False, False],
                    'error_count': 0
                },
                'round': 1,
                'bot_turn': 0,
            }
        }
    }
    sock.send_json(get_move)
    player_handle_request(client_sock, team)

    assert sock.recv_json() == {
        '__uuid__': _uuid,
        '__return__': {
            "move": [1, 1],
            "say": None
        }
    }

    exit_msg = {
        '__uuid__': uuid.uuid4().__str__(),
        '__action__': "exit",
        '__data__': {}
    }
    sock.send_json(exit_msg)

    assert res[0] == True



import pytest
from unittest.mock import MagicMock

import re
from textwrap import dedent

from pelita import tournament
from pelita.tournament import knockout_mode, roundrobin
from pelita.tournament.knockout_mode import Team, Match, Bye


def test_match_id():
    assert str(tournament.MatchID()) == 'round1-match01'
    assert str(tournament.MatchID(round=1)) == 'round1-match01'
    assert str(tournament.MatchID(round=2)) == 'round2-match01'
    assert str(tournament.MatchID(round=1, match=1)) == 'round1-match01'
    assert str(tournament.MatchID(round=1, match=2, match_repeat=1)) == 'round1-match02'
    assert str(tournament.MatchID(round=1, match=2, match_repeat=2)) == 'round1-match02-repeat2'

    match_id = tournament.MatchID(round=1, match=2, match_repeat=2)
    match_id.next_round()
    assert match_id == tournament.MatchID(round=2, match=1, match_repeat=1)

    match_id = tournament.MatchID(round=1, match=2, match_repeat=2)
    match_id.next_match()
    assert match_id == tournament.MatchID(round=1, match=3, match_repeat=1)

    match_id = tournament.MatchID(round=1, match=2, match_repeat=2)
    match_id.next_repeat()
    assert match_id == tournament.MatchID(round=1, match=2, match_repeat=3)


class TestKoMode:
    def test_sort_ranks(self):
        sort_ranks = knockout_mode.sort_ranks
        assert sort_ranks([0, 1, 2, 3, 4, 5, 6]) == [0, 5, 1, 4, 2, 3, 6]
        assert sort_ranks([0, 1, 2, 3, 4, 5]) == [0, 5, 1, 4, 2, 3]
        assert sort_ranks([0, 1, 2, 3, 4]) == [0, 3, 1, 2, 4]
        assert sort_ranks([0, 1, 2, 3]) == [0, 3, 1, 2]
        assert sort_ranks([0, 1, 2]) == [0, 1, 2]
        assert sort_ranks([0, 1]) == [0, 1]
        assert sort_ranks([0]) == [0]
        assert sort_ranks([]) == []

        assert sort_ranks([0, 1, 2, 3, 4, 5, 6], bonusmatch=True) == [0, 5, 1, 4, 2, 3, 6]
        assert sort_ranks([0, 1, 2, 3, 4, 5], bonusmatch=True) == [0, 3, 1, 2, 4, 5]
        assert sort_ranks([0, 1, 2, 3, 4], bonusmatch=True) == [0, 3, 1, 2, 4]
        assert sort_ranks([0, 1, 2, 3], bonusmatch=True) == [0, 1, 2, 3]
        assert sort_ranks([0, 1, 2], bonusmatch=True) == [0, 1, 2]
        assert sort_ranks([0, 1], bonusmatch=True) == [0, 1]
        assert sort_ranks([0], bonusmatch=True) == [0]
        assert sort_ranks([], bonusmatch=True) == []


    def test_prepared_matches(self):
        with pytest.raises(ValueError):
            none = knockout_mode.prepare_matches([])
        with pytest.raises(ValueError):
            none = knockout_mode.prepare_matches([], bonusmatch=True)

        single = knockout_mode.prepare_matches([1])
        assert single == knockout_mode.Team(name=1)
        single = knockout_mode.prepare_matches([1], bonusmatch=True)
        assert single == knockout_mode.Team(name=1)

        pair = knockout_mode.prepare_matches([1,2])
        assert pair == Match(t1=Team(name=1), t2=Team(name=2))
        pair = knockout_mode.prepare_matches([1,2], bonusmatch=True)
        assert pair == Match(t1=Team(name=1), t2=Team(name=2))

        triple = knockout_mode.prepare_matches([1,2,3])
        assert triple == Match(t1=Match(t1=Team(name=1), t2=Team(name=2)), t2=Bye(team=Team(name=3)))
        triple = knockout_mode.prepare_matches([1,2,3], bonusmatch=True)
        assert triple == Match(t1=Match(t1=Team(name=1), t2=Team(name=2)), t2=Bye(team=Team(name=3)))

        matches = knockout_mode.prepare_matches([1,2,3,4])
        outcome = Match(t1=Match(t1=Team(name=1), t2=Team(name=4)), t2=Match(t1=Team(name=2), t2=Team(name=3)))
        assert matches == outcome
        matches = knockout_mode.prepare_matches([1,2,3,4], bonusmatch=True)
        outcome = Match(t1=Match(t1=Match(t1=Team(name=1), t2=Team(name=2)), t2=Bye(team=Team(name=3))), t2=Bye(team=Bye(team=Team(name=4))))
        assert matches == outcome

        matches = knockout_mode.prepare_matches([1,2,3,4,5])
        outcome = Match(t1=Match(t1=Match(t1=Team(name=1), t2=Team(name=4)), t2=Match(t1=Team(name=2), t2=Team(name=3))), t2=Bye(team=Bye(team=Team(name=5))))
        assert matches == outcome
        matches = knockout_mode.prepare_matches([1,2,3,4,5], bonusmatch=True)
        outcome = Match(t1=Match(t1=Match(t1=Team(name=1), t2=Team(name=4)), t2=Match(t1=Team(name=2), t2=Team(name=3))), t2=Bye(team=Bye(team=Team(name=5))))
        assert matches == outcome

        matches = knockout_mode.prepare_matches([1,2,3,4,5,6])
        outcome = Match(t1=Match(t1=Match(t1=Team(name=1), t2=Team(name=6)), t2=Match(t1=Team(name=2), t2=Team(name=5))), t2=Bye(team=Match(t1=Team(name=3), t2=Team(name=4))))
        assert matches == outcome
        matches = knockout_mode.prepare_matches([1,2,3,4,5,6], bonusmatch=True)
        outcome = Match(t1=Match(t1=Match(t1=Match(t1=Team(name=1), t2=Team(name=4)), t2=Match(t1=Team(name=2), t2=Team(name=3))), t2=Bye(team=Bye(team=Team(name=5)))), t2=Bye(team=Bye(team=Bye(team=Team(name=6)))))
        assert matches == outcome


@pytest.mark.parametrize('teams, bonusmatch, check_output', [
    ([1, 2, 3, 4, 5], True, """
        1 ┐
          ├─ ??? ┐
        4 ┘      │
                 ├─ ??? ┐
        2 ┐      │      │  ┏━━━━━┓
          ├─ ??? ┘      ├──┨ ??? ┃
        3 ┘             │  ┗━━━━━┛
                        │
        5 ──────────────┘
        """),
    ([1, 2, 3, 4, 5, 6], False, """
        1 ┐
          ├─ ??? ┐
        6 ┘      │
                 ├─ ??? ┐
        2 ┐      │      │
          ├─ ??? ┘      │  ┏━━━━━┓
        5 ┘             ├──┨ ??? ┃
                        │  ┗━━━━━┛
        3 ┐             │
          ├─ ??? ───────┘
        4 ┘
        """),
    ([1, 2, 3, 4, 5, 6], True, """
        1 ┐
          ├─ ??? ┐
        4 ┘      │
                 ├─ ??? ┐
        2 ┐      │      │
          ├─ ??? ┘      ├─ ??? ┐
        3 ┘             │      │  ┏━━━━━┓
                        │      ├──┨ ??? ┃
        5 ──────────────┘      │  ┗━━━━━┛
                               │
        6 ─────────────────────┘
        """),
])
def test_knockout_output(teams, bonusmatch, check_output):
    matches = knockout_mode.prepare_matches(teams, bonusmatch=bonusmatch)
    printed = knockout_mode.print_knockout(matches)
    # remove the whitespace to the right of the printed lines
    printed = re.sub(r'\s+$', '', printed, flags=re.MULTILINE)
    assert dedent(printed).strip() == dedent(check_output).strip()


class TestRoundRobin:
    def test_all_matches_included(self):
        data = [
            ([], []),
            ([1], []),
            ([1, 2], [(1, 2)]),
            ([1, 2, 3], [(1, 2), (1, 3), (2, 3)]),
            ([1, 2, 3, 4], [(1, 2), (1, 3), (1, 4), (2, 3), (2, 4), (3, 4)]),
            ([0, 1, 2, 3, 4], [(0, 1), (0, 2), (0, 3), (0, 4), (1, 2), (1, 3), (1, 4), (2, 3), (2, 4), (3, 4)])
        ]

        def tuple_sort(t):
            # sort a tuple
            return tuple(sorted(list(t)))

        for input, output in data:
            matchplan = roundrobin.create_matchplan(input)
            matchplan_sorted = sorted([tuple_sort(x) for x in matchplan])
            output = sorted([tuple_sort(x) for x in output])
            assert matchplan_sorted == output

    @pytest.mark.parametrize('teams',  [
            [0, 1, 2, 3, 4],
            [0, 1, 2, 3, 4, 5],
            [0, 1, 2, 3, 4, 5, 6],
            [0, 1, 2, 3, 4, 5, 6, 7]
        ])
    def test_no_team_plays_two_games_in_a_row(self, teams):
        matches = roundrobin.create_matchplan(teams)
        prev = {}
        for match in matches:
            current = set(match)
            assert current.isdisjoint(prev)
            prev = current

    @pytest.mark.parametrize('teams',  [
            [0, 1],
            [0, 1, 2],
            [0, 1, 2, 3],
            [0, 1, 2, 3, 4],
            [0, 1, 2, 3, 4, 5],
            [0, 1, 2, 3, 4, 5, 6],
            [0, 1, 2, 3, 4, 5, 6, 7],
        ])
    def test_team_blue_red_are_similar(self, teams):
        # test that each team plays a similar amount as blue and red
        matchplan = roundrobin.create_matchplan(teams)
        from collections import Counter
        blue_count = Counter()
        red_count = Counter()
        for blue, red in matchplan:
            blue_count[blue] += 1
            red_count[red] += 1
        for team in teams:
            assert -1 <= blue_count[team] - red_count[team] <= 1, matchplan

    @pytest.mark.parametrize('list, fixed, outcome',  [
            ([0], 0, [0]),
            ([0, 1], 0, [0, 1]),
            ([0, 1], 1, [0, 1]),
            ([0, 1, 2], 0, [0, 2, 1]),
            ([0, 1, 2], 1, [2, 1, 0]),
            ([0, 1, 2], 2, [1, 0, 2]),
            ([0, 1, 2, 3], 2, [1, 3, 2, 0]),
            ([0, 1, 2, 3, 4], 3, [1, 2, 4, 3, 0]),
            ([0, 1, 2, 3, 4, 5], 2, [1, 3, 2, 4, 5, 0])
        ])
    def test_rotate_with_fixed(self, list, fixed, outcome):
        assert roundrobin.rotate_with_fixed(list, fixed) == outcome


### ASSERTIONS:
# There must be exactly one game_state with finished=True

class TestSingleMatch:
    def test_play_game_with_config(self):
        config = MagicMock()
        config.rounds = 200
        config.team_spec = lambda x: x
        config.team_group = lambda x: x
        config.viewer = 'ascii'
        config.size = 'small'
        config.tournament_log_folder = None

        teams = ["pelita/player/StoppingPlayer", "pelita/player/StoppingPlayer"]
        (state, stdout, stderr) = tournament.play_game_with_config(config, teams)
        assert state['whowins'] == 2

        config.rounds = 200
        config.team_spec = lambda x: x
        config.viewer = 'ascii'
        teams = ["pelita/player/SmartEatingPlayer", "pelita/player/StoppingPlayer"]
        (state, stdout, stderr) = tournament.play_game_with_config(config, teams)
        print(state)
        assert state['whowins'] == 0

        config.rounds = 200
        config.team_spec = lambda x: x
        config.viewer = 'ascii'
        teams = ["pelita/player/StoppingPlayer", "pelita/player/SmartEatingPlayer"]
        (state, stdout, stderr) = tournament.play_game_with_config(config, teams)
        assert state['whowins'] == 1

    def test_start_match(self):
        stdout = []

        def mock_print(str="", *args, **kwargs):
            print(str)
            stdout.append(str)

        teams = {
            "first_id": "pelita/player/StoppingPlayer",
            "second_id": "pelita/player/SmartEatingPlayer",
        }

        config = MagicMock()
        config.rounds = 300
        config.team_spec = lambda x: teams[x]
        config.team_name = lambda x: teams[x]
        config.team_group = lambda x: x
        config.viewer = 'ascii'
        config.size = 'small'
        config.print = mock_print
        config.tournament_log_folder = None

        team_ids = ["first_id", "first_id"]
        result = tournament.start_match(config, team_ids)
        assert result == False
        assert stdout[-1] == '‘pelita/player/StoppingPlayer’ and ‘pelita/player/StoppingPlayer’ had a draw.'

        team_ids = ["second_id", "first_id"]
        result = tournament.start_match(config, team_ids)
        assert result == "second_id"
        assert stdout[-1] == '‘pelita/player/SmartEatingPlayer’ wins'

        team_ids = ["first_id", "second_id"]
        result = tournament.start_match(config, team_ids)
        assert result == "second_id"
        assert stdout[-1] == '‘pelita/player/SmartEatingPlayer’ wins'


    def test_deathmatch(self):
        stdout = []

        def mock_print(str="", *args, **kwargs):
            stdout.append(str)

        teams = {
            "first_id": "pelita/player/StoppingPlayer",
            "second_id": "pelita/player/StoppingPlayer",
        }

        config = MagicMock()
        config.rounds = 200
        config.teams = teams
        config.team_spec = lambda x: teams[x]
        config.team_name = lambda x: teams[x]
        config.team_group = lambda x: x
        config.viewer = 'ascii'
        config.size = 'small'
        config.print = mock_print
        config.tournament_log_folder = None

        result = tournament.start_deathmatch(config, *teams.keys())
        assert result is not None
        assert result in ["first_id", "second_id"]


class TestTournament:
    def test_team_id_check(self):
        with pytest.raises(ValueError):
            tournament.create_team_id(1, 0)
        with pytest.raises(ValueError):
            tournament.create_team_id("#abc", 0)
        with pytest.raises(ValueError):
            tournament.create_team_id("", 0)
        assert tournament.create_team_id(None, 3) == "#3"
        assert tournament.create_team_id("abc", 3) == "abc"

    def test_tournament_winner(self):
        stdout = []

        def mock_print(str="", *args, **kwargs):
            stdout.append(str)

        c = {
            "location": None,
            "date": None,
            "bonusmatch": None,
            "teams": [
                {"id": "group0", "spec": "pelita/player/StoppingPlayer", "members": []},
                {"id": "group1", "spec": "pelita/player/SmartEatingPlayer", "members": []},
                {"id": "group2", "spec": "pelita/player/StoppingPlayer", "members": []},
                {"id": "group3", "spec": "pelita/player/StoppingPlayer", "members": []},
                {"id": "group4", "spec": "pelita/player/StoppingPlayer", "members": []},
            ],
            "size": "small",
        }
        config = tournament.Config(c)
        config.print = mock_print
        config.viewer = 'null'
        config.state = None
        config.tournament_log_folder = None

        # group1 should win
        assert "group1" == tournament.start_match(config, ["group0", "group1"])
        assert "group1" == tournament.start_match(config, ["group1", "group0"])
        assert False == tournament.start_match(config, ["group0", "group0"])

        tournament.present_teams(config)

        state = tournament.State(config)
        rr_ranking = tournament.play_round1(config, state)

        if config.bonusmatch:
            sorted_ranking = knockout_mode.sort_ranks(rr_ranking[:-1]) + [rr_ranking[-1]]
        else:
            sorted_ranking = knockout_mode.sort_ranks(rr_ranking)

        winner = tournament.play_round2(config, sorted_ranking, state)
        assert winner == 'group1'




import networkx

from pelita.utils import walls_to_graph


def food_eating_player(bot, state):
    if 'graph' not in state:
        # first turn, first round
        state['graph'] = walls_to_graph(bot.walls)

    if bot.turn not in state:
        state[bot.turn] = { 'next_food': None }

    # check if food is still there for us to eat
    if (state[bot.turn]['next_food'] is None
        or state[bot.turn]['next_food'] not in bot.enemy[0].food):
        if not bot.enemy[0].food:
            # all food has been eaten? ok. I’ll stop
            # NB: We should never land here; the game will be over instead
            next_pos = bot.position
            return next_pos

        state[bot.turn]['next_food'] = bot.random.choice(bot.enemy[0].food)

    # the first position in the shortest path is always bot.position
    next_pos = networkx.shortest_path(state['graph'], bot.position, state[bot.turn]['next_food'])[1]

    return next_pos


TEAM_NAME = "Food Eating Players"
move = food_eating_player



#!/usr/bin/env python3

import argparse
import datetime
import itertools
from pathlib import Path
import random
import re
import shlex
import sys

import shutil
import yaml

from .. import tournament
from .script_utils import start_logging

def firstNN(*args):
    """
    Return the first argument not None.

    Example
    -------
        >>> firstNN(None, False, True)
        False
        >>> firstNN(True, False, True)
        True
        >>> firstNN(None, None, True)
        True
        >>> firstNN(None, 2, True)
        2
        >>> firstNN(None, None, None)
        None
        >>> firstNN()
        None
    """
    return next(filter(lambda x: x is not None, args), None)


def shlex_unsplit(cmd):
    """
    Translates a list of command arguments into bash-like ‘human’ readable form.
    Pseudo-reverses shlex.split()

    Example
    -------
        >>> shlex_unsplit(["command", "-f", "Hello World"])
        "command -f 'Hello World'"

    Parameters
    ----------
    cmd : list of string
        command + parameter list

    Returns
    -------
    string
    """
    return " ".join(shlex.quote(arg) for arg in cmd)


def create_directory(prefix):
    for suffix in itertools.count(0):
        path = Path('{}-{:02d}'.format(prefix, suffix))
        try:
            path.mkdir()
        except FileExistsError:
            pass
        else:
            break
    return path

def input_choice(text, choices, vars):
    selected = ""
    while not selected or selected not in vars:
        print(text)
        for choice in choices:
            print(choice)
        selected = input().strip()
    return selected


def autoconf_sound():
    res = {}
    which_say = shutil.which("say")
    if which_say:
        res["say"] = [which_say, "-f"]
    which_flite = shutil.which("flite")
    if which_flite:
        res["flite"] = [which_flite]
    return res


def setup():
    config = {}
    print("Where should the next tournament be?")
    config['location'] = input()
    print("When should the next tournament be? (ex. {year})".format(year=datetime.datetime.now().year))
    config['date'] = input()
    print("What is the name of the tournament host?")
    config['host'] = input()
    print("What is the local greeting?")
    config['greeting'] = input()
    print("What is the local farewell?")
    config['farewell'] = input()

    while True:
        sound = autoconf_sound()
        sound_options = [
            "(e)nter manually",
        ]
        keys = "e"

        try:
            which_say = sound["say"]
            sound_options.append("(s)ay ({})".format(sound["say"]))
            keys += "s"
        except KeyError:
            pass

        try:
            which_say = sound["flite"]
            sound_options.append("(f)flite ({})".format(sound["flite"]))
            keys += "f"
        except KeyError:
            pass

        sound_options.append("(n)o sound")
        keys += "n"
        sound_options.append("(r)etry search")
        keys += "r"

        res = input_choice("Enable sound?", sound_options, keys)
        if res == "n":
            config["speak"] = False
            break
        elif res == "e":
            print("Please enter the location of the sound-giving binary:")
            sound_path = input()
        elif res == "s":
            sound_path = shlex_unsplit(sound["say"])
        elif res == "f":
            sound_path = shlex_unsplit(sound["flite"])
        else:
            continue

        print("Now trying to speak:")
        config["speak"] = True
        config['speaker'] = sound_path
        # must set a few dummy variables
        config['teams'] = []
        config['bonusmatch'] = []

        greeting = config.get("greeting", "Hello")
        farewell = config.get("farewell", "Good evening")
        host = config.get("host", "host")

        tournament.Config(config).say(f"{greeting} {host}. {farewell} {host}.")
        success = input_choice("Did you hear any sound? (y/n)", [], "yn")
        if success == "y":
            del config["teams"]
            del config["bonusmatch"]
            break

    print("Specify the folder where we should look for teams (or none)")
    folder = input().strip()
    if folder:
        try:
            subfolders = [x.as_posix() for x in Path(folder).iterdir() if x.is_dir()
                                                                       and not x.name.startswith('.')
                                                                       and not x.name.startswith('_')]
            config["teams"] = [{ 'spec': folder, 'members': []} for folder in subfolders]
        except FileNotFoundError:
            print("Invalid path: {}".format(folder))

    res = input_choice("Should a bonus match be played? (y/n/i)", [], "yni")
    if res == "y":
        config['bonusmatch'] = True
    elif res == "n":
        config['bonusmatch'] = False

    def escape(str):
        return "-" + re.sub(r'[\W]', '_', str) if str else ""

    file_name = "tournament{location}{year}.yaml".format(location=escape(config["location"]),
                                                         year=escape(config["date"]))

    print("Writing to: {file_name}".format(file_name=file_name))
    with open(file_name, "x") as f:
        yaml.dump(config, f, default_flow_style=False)


def main():
    parser = argparse.ArgumentParser(description='Run a tournament',
                                     add_help=False,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser._positionals = parser.add_argument_group('Arguments')
    parser._optionals = parser.add_argument_group('Options')
    parser.add_argument('--help', '-h',
                        help='show this help message and exit',
                        action='store_true')

    parser.add_argument('--speak', dest='speak', action='store_true', help='speak loudly every message on stdout')
    parser.add_argument('--no-speak', dest='speak', action='store_false', help='do not speak every message on stdout')
    parser.set_defaults(speak=None)

    parser.add_argument('--speaker', help='tool to say stuff', type=str)

    parser.add_argument('--rounds', '-r',
                        help='maximum number of rounds to play per match',
                        type=int)
    parser.add_argument('--viewer', '-v',
                        type=str, help='the pelita viewer to use (default: tk)')
    parser.add_argument('--config', help='tournament data',
                        metavar="CONFIG_YAML", default="tournament.yaml")

    parser.add_argument('--interactive', dest='interactive', action='store_true', help='do not ask before proceeding')
    parser.add_argument('--non-interactive', dest='interactive', action='store_false', help='do not ask before proceeding')
    parser.set_defaults(interactive=None)

    parser.add_argument('--setup', action='store_true')

    parser.add_argument('--state', help='store state',
                        metavar="STATEFILE", default='state.yaml')
    parser.add_argument('--load-state', help='load state from file',
                        action='store_true')
    parser.add_argument('--no-log', help='do not store the log data',
                        action='store_true')
    parser.add_argument('--dry-run', help='do not actually play',
                        action='store_true')

    args = parser.parse_args()
    if args.help:
        parser.print_help()
        sys.exit(0)
    elif args.setup:
        setup()
        sys.exit(0)

    try:
        with open(args.config) as f:
            config_data = yaml.load(f, Loader=yaml.FullLoader)
            config_data['viewer'] = args.viewer or config_data.get('viewer', 'tk')
            config_data['interactive'] = firstNN(args.interactive, config_data.get('interactive'), True)
            config_data['statefile'] = args.state
            config_data['speak'] = firstNN(args.speak, config_data.get('speak'))
            config_data['speaker'] = args.speaker or config_data.get('speaker')

            config = tournament.Config(config_data)
    except FileNotFoundError:
        print("‘{}’ not found. Create a new tournament with ‘--setup’.".format(args.config))
        sys.exit(1)

    if not args.no_log:
        # create a directory for the dumps
        def escape(s):
            return "-" + re.sub(r'[\W]', '_', str(s)) if s else ""

        storage_folder = create_directory(f'./store{escape(config.location)}{escape(config.date)}')

        config.tournament_log_folder = storage_folder
        config.tournament_log_file = storage_folder / 'tournament.out'

        start_logging(storage_folder / 'tournament.log')

    if args.rounds:
        config.rounds = args.rounds

    if Path(args.state).is_file():
        if not args.load_state:
            config.print(f"Found state file in {args.state}. Restore with --load-state. Aborting.")
            sys.exit(-1)
        else:
            state = tournament.State.load(config, args.state)
    else:
        state = tournament.State(config)

    random.seed(config.seed)

    if state.round1['played']:
        # We have already played one match. Do not speak the introduction.
        old_speak = config.speak
        config.speak = False
        tournament.present_teams(config)
        config.speak = old_speak
    else:
        tournament.present_teams(config)

    rr_ranking = tournament.play_round1(config, state)
    state.round2["round_robin_ranking"] = rr_ranking
    state.save(args.state)

    winner = tournament.play_round2(config, rr_ranking, state)

    config.print('The winner of the %s Pelita tournament is...' % config.location, wait=2, end=" ")
    config.print('{team_group}: {team_name}. Congratulations'.format(
        team_group=config.team_group(winner),
        team_name=config.team_name(winner)), wait=2)

    farewell = config.farewell
    host = config.host

    config.print(f'{farewell} {host}. It was a pleasure to serve you.', wait=2)

    config.print('*** Please remember to copy the log files from {}. ***'.format(config.tournament_log_folder), speak=False)

if __name__ == '__main__':
    main()



def random_player(bot, state):
    return bot.random.choice(bot.legal_positions)

def nq_random_player(bot, state):
    """ Not-Quite-RandomPlayer that will move randomly but not stop or reverse. """

    legal_positions = bot.legal_positions[:]
    # Remove stop
    try:
        legal_positions.remove(bot.position)
    except ValueError:
        pass
    # now remove the move that would lead to the previous_position
    # unless there is no where else to go.
    if len(legal_positions) > 1:
        if len(bot.track) >= 2:
            try:
                legal_positions.remove(bot.track[-2])
            except ValueError:
                # if we did not move in the last round,
                # there will be nothing left to delete
                pass
    # just in case, there is really no way to go to:
    if not legal_positions:
        return bot.position
    # and select a move at random
    return bot.random.choice(legal_positions)


TEAM_NAME = "Random Players"
move = nq_random_player




from pelita.player import food_eating_player

def smart_eating_player(bot, state):
    # food eating player but won’t do kamikaze (although a sufficently smart
    # enemy will be able to kill the bot in its next turn as it doesn’t flee)
    next_pos = food_eating_player(bot, state)

    dangerous_enemy_pos = [enemy.position for enemy in bot.enemy if enemy.position in enemy.homezone]

    # check, if the next_pos has an enemy on it
    if next_pos in dangerous_enemy_pos:
        # whoops, better wait this round and take another food next time
        state[bot.turn]['next_food'] = None
        return bot.position

    return next_pos


TEAM_NAME = "Smart Eating Players"
move = smart_eating_player



import logging
import time

import zmq

try:
    import tkinter
except ImportError:
    # if we are on linux, we suggest to install the python3-tk package
    import platform
    if 'Linux' in platform.platform():
        message = '\nYour Python installation is missing tkinter\n\n'
        if platform.freedesktop_os_release()['ID'] == 'debian':
            message += 'Please install the python3-tk package'
        else:
            message += 'Please install the python3-tkinter package'
        raise Exception(message)

import tkinter.font

from ..game import next_round_turn
from ..team import _ensure_list_tuples
from .tk_sprites import (
    BotSprite,
    Food,
    Wall,
    Arrow,
    RED,
    BLUE,
    YELLOW,
    GREY,
    LIGHT_GREY,
    SELECTED,
    BROWN,
    LIGHT_BLUE,
    LIGHT_RED,
    STRONG_BLUE,
    STRONG_RED,
)
from .tk_utils import wm_delete_window_handler
from .. import layout

_logger = logging.getLogger(__name__)


def guess_size(display_string, bounding_width, bounding_height, rel_size=0):
    no_lines = display_string.count("\n") + 1
    size_guess = bounding_height // ((3-rel_size) * no_lines)
    font = tkinter.font.Font(size=size_guess)
    text_width = font.measure(display_string)
    if text_width > bounding_width:
        font_size = size_guess * bounding_width // text_width
    else:
        font_size = size_guess
    return font_size

class MeshGraph:
    """ A `MeshGraph` is a structure of `mesh_width` * `mesh_height` rectangles,
    covering an area of `screen_width`, `screen_height`.
    """
    def __init__(self, mesh_width, mesh_height, screen_width, screen_height):
        self.mesh_width = mesh_width
        self.mesh_height = mesh_height
        self.screen_height = screen_height
        self.screen_width = screen_width
        self.padding = 10

    @property
    def rect_width(self):
        """ The width of a single field.
        """
        # we have to adjust by one pixel for the border
        width = float(self.screen_width - 1 - 2 * self.padding) / self.mesh_width
        # if the UI is not initialized yet (or just really really small) this may
        # result in a negative value. Ensure that it is always positive or zero.
        if width < 0:
            width = 0.0
        return width

    @property
    def rect_height(self):
        """ The height of a single field.
        """
        # we have to adjust by one pixel for the border
        height = float(self.screen_height - 1 - 2 * self.padding) / self.mesh_height
        # if the UI is not initialized yet (or just really really small) this may
        # result in a negative value. Ensure that it is always positive or zero.
        if height < 0:
            height = 0.0
        return height

    @property
    def half_scale_x(self):
        return self.rect_width / 2.0

    @property
    def half_scale_y(self):
        return self.rect_height / 2.0

    def mesh_trafo(self, mesh_x, mesh_y):
        return Trafo(self, mesh_x, mesh_y)

    def mesh_to_screen(self, mesh, coords):
        mesh_x, mesh_y = mesh
        coords_x, coords_y = coords

        real_x = self.mesh_to_screen_x(mesh_x, coords_x)
        real_y = self.mesh_to_screen_y(mesh_y, coords_y)
        return (real_x, real_y)

    def mesh_to_screen_x(self, mesh_x, model_x):
        # coords are between -1 and +1: shift on [0, 1]
        trafo_x = (model_x + 1.0) / 2.0

        real_x = self.rect_width * (mesh_x + trafo_x) + self.padding
        return real_x

    def mesh_to_screen_y(self, mesh_y, model_y):
        # coords are between -1 and +1: shift on [0, 1]
        trafo_y = (model_y + 1.0) / 2.0

        real_y = self.rect_height * (mesh_y + trafo_y) + self.padding
        return real_y

    def screen_to_mesh_coord(self, screen_x, screen_y):
        # returns the mesh coordinate of the selected screen coordinate
        # or None, when it is outside of the mesh

        x = int((screen_x - self.padding) / self.rect_width)
        y = int((screen_y - self.padding) / self.rect_height)

        if not 0 <= x < self.mesh_width or not 0 <= y < self.mesh_height:
            return None
        return (x, y)

    def __repr__(self):
        return "MeshGraph(%d, %d, %d, %d)" % (self.mesh_width, self.mesh_height,
                                              self.screen_width, self.screen_height)

class Trafo:
    def __init__(self, mesh_graph, mesh_x, mesh_y):
        self.mesh_graph = mesh_graph
        self.mesh_x = mesh_x
        self.mesh_y = mesh_y

    def screen_x(self, model_x):
        return self.mesh_graph.mesh_to_screen_x(self.mesh_x, model_x)

    def screen_y(self, model_y):
        return self.mesh_graph.mesh_to_screen_y(self.mesh_y, model_y)

    def screen(self, model_x, model_y):
        return self.mesh_graph.mesh_to_screen((self.mesh_x, self.mesh_y), (model_x, model_y))

class UI:
    pass

class TkApplication:
    def __init__(self, window, controller_address=None,
                 geometry=None, delay=1, stop_after=None, fullscreen=False):
        self.window = window
        self.window.configure(background="white")

        self.context = zmq.Context()

        if controller_address:
            self.controller_socket = self.context.socket(zmq.DEALER)
            self.controller_socket.connect(controller_address)
        else:
            self.controller_socket = None

        self.window.title("Pelita")

        self.game_finish_overlay = lambda: None

        self.mesh_graph = None
        self.geometry = geometry
        self.fullscreen = fullscreen
        self._fullscreen_enabled = fullscreen

        self._default_font = tkinter.font.nametofont("TkDefaultFont")
        self._default_font_size = self._default_font.cget('size')

        self.size_changed = True

        self._grid_enabled = False

        self._times = []
        self._fps = None
        self.selected = None

        self.game_uuid = None
        self.bot_sprites = {}
        self.shadow_sprites = {}

        self._universe = None
        self._game_state = None

        self.ui = UI()

        self.ui.header_canvas = tkinter.Canvas(window, height=50)
        self.ui.header_canvas.config(background="white", bd=0, highlightthickness=0, relief='ridge')

        self.ui.sub_header = tkinter.Frame(window, height=25)
        self.ui.sub_header.config(background="white")

        self.ui.status_canvas = tkinter.Frame(window, height=25)
        self.ui.status_canvas.config(background="white")

        self.ui.game_canvas = tkinter.Canvas(window)
        self.ui.game_canvas.config(background="white", bd=0, highlightthickness=0, relief='ridge')
        self.ui.game_canvas.bind('<Configure>', lambda e: window.after_idle(self.update))
        self.ui.game_canvas.bind('<Button-1>', self.on_click)

        self.ui.status_00 = tkinter.Frame(self.ui.status_canvas, background="white", padx=5, pady=0)
        self.ui.status_01 = tkinter.Frame(self.ui.status_canvas, background="white", padx=5, pady=0)
        self.ui.status_02 = tkinter.Frame(self.ui.status_canvas, background="white", padx=5, pady=0)
        self.ui.status_10 = tkinter.Frame(self.ui.status_canvas, background="white", padx=5, pady=0)
        self.ui.status_11 = tkinter.Frame(self.ui.status_canvas, background="white", padx=5, pady=0)
        self.ui.status_12 = tkinter.Frame(self.ui.status_canvas, background="white", padx=5, pady=0)
        self.ui.status_20 = tkinter.Frame(self.ui.status_canvas, background="white", padx=5, pady=0)
        self.ui.status_21 = tkinter.Frame(self.ui.status_canvas, background="white", padx=5, pady=0)
        self.ui.status_22 = tkinter.Frame(self.ui.status_canvas, background="white", padx=5, pady=0)

        self.ui.status_00.grid(row=0, column=0, sticky="W")
        self.ui.status_01.grid(row=0, column=1, sticky="WE")
        self.ui.status_02.grid(row=0, column=2, sticky="E")
        self.ui.status_10.grid(row=1, column=0, sticky="W")
        self.ui.status_11.grid(row=1, column=1, columnspan=2, sticky="E")
#        self.ui.status_12.grid(row=1, column=2, sticky="E")
        self.ui.status_20.grid(row=2, column=0, sticky="W")
        self.ui.status_21.grid(row=2, column=1, sticky="W")
        self.ui.status_22.grid(row=2, column=2, sticky="E")

        self.ui.status_canvas.grid_columnconfigure(0, weight=1, uniform='status')
        self.ui.status_canvas.grid_columnconfigure(1, weight=1, uniform='status')
        self.ui.status_canvas.grid_columnconfigure(2, weight=1, uniform='status')

        self.ui.button_game_speed_slower = tkinter.Button(self.ui.status_10,
            foreground="black",
            background="white",
            justify=tkinter.CENTER,
            text="slower",
            padx=12,
            command=self.delay_inc)
        self.ui.button_game_speed_slower.pack(side=tkinter.LEFT)

        self.ui.button_game_speed_faster = tkinter.Button(self.ui.status_10,
            foreground="black",
            background="white",
            justify=tkinter.CENTER,
            text="faster",
            padx=12,
            command=self.delay_dec)
        self.ui.button_game_speed_faster.pack(side=tkinter.LEFT)

        self._check_speed_button_state()

        self.ui.button_game_toggle_grid = tkinter.Button(self.ui.status_10,
            foreground="black",
            background="white",
            justify=tkinter.CENTER,
            padx=12,
            command=self.toggle_grid)
        self.ui.button_game_toggle_grid.pack(side=tkinter.LEFT)

        self._check_grid_toggle_state()

        self.ui.status_fps_info = tkinter.Label(self.ui.status_20,
            text="",
            font=(None, 8),
            foreground="black",
            background="white",
            justify=tkinter.CENTER)
        self.ui.status_fps_info.pack(side=tkinter.LEFT)

        self.ui.status_selected = tkinter.Label(self.ui.status_22,
            text="",
            font=(None, 8),
            foreground="black",
            background="white",
            justify=tkinter.CENTER)
        self.ui.status_selected.pack(side=tkinter.RIGHT)

        tkinter.Button(self.ui.status_00,
                       foreground="black",
                       background="white",
                       justify=tkinter.CENTER,
                       text="PLAY/PAUSE",
                       padx=12,
                       command=self.toggle_running).pack(side=tkinter.LEFT, expand=tkinter.YES)

        tkinter.Button(self.ui.status_00,
                       foreground="black",
                       background="white",
                       justify=tkinter.CENTER,
                       text="STEP",
                       padx=12,
                       command=self.request_step).pack(side=tkinter.LEFT, expand=tkinter.YES)

        tkinter.Button(self.ui.status_00,
                       foreground="black",
                       background="white",
                       justify=tkinter.CENTER,
                       text="ROUND",
                       padx=12,
                       command=self.request_round).pack(side=tkinter.LEFT, expand=tkinter.YES)

        tkinter.Button(self.ui.status_01,
                       foreground="black",
                       background="white",
                       justify=tkinter.CENTER,
                       text="QUIT",
                       width=30,
                       padx=12,
                       command=self.quit).pack(side=tkinter.TOP, fill=tkinter.BOTH, anchor=tkinter.CENTER)

        # Add circles to show which bot is active.
        self.ui.bot_traffic_lights_canvas = tkinter.Canvas(self.ui.status_02, height=25, width=44,
                                                           background="white", bd=0, highlightthickness=0,
                                                           relief='ridge')
        self.ui.bot_traffic_lights_canvas.pack(side=tkinter.LEFT)
        self.ui.bot_traffic_lights_c = []
        for idx in range(4):
            spacing = 10
            x0, y0 = 2.5 + idx * spacing, 13
            circle = self.ui.bot_traffic_lights_canvas.create_text(x0, y0, text=layout.BOT_I2N[idx], font=(None, 11), fill="black")
            self.ui.bot_traffic_lights_c.append(circle)

        self.ui.status_round_info = tkinter.Label(self.ui.status_02, text="", background="white")
        self.ui.status_round_info.pack(side=tkinter.LEFT)

        self.ui.status_layout_info = tkinter.Label(self.ui.status_11, text="", background="white")
        self.ui.status_layout_info.pack(side=tkinter.LEFT)

        self.ui.header_canvas.pack(side=tkinter.TOP, fill=tkinter.BOTH)
#        self.ui.sub_header.pack(side=tkinter.TOP, fill=tkinter.BOTH)
        self.ui.status_canvas.pack(side=tkinter.BOTTOM, fill=tkinter.BOTH)
        self.ui.game_canvas.pack(side=tkinter.TOP, fill=tkinter.BOTH, expand=tkinter.YES)#fill=tkinter.BOTH, expand=tkinter.YES)

        self._min_delay = 1
        self._delay = delay
        self._stop_after = stop_after
        self._stop_after_delay = delay
        if self._stop_after is not None:
            self._delay = self._min_delay

        self._check_speed_button_state()

        self._observed_steps = set()

        self.running = True

        self.window.bind('f', lambda event: self.toggle_fullscreen())
        self.window.bind('q', lambda event: self.quit())
        self.window.bind('<numbersign>', lambda event: self.toggle_grid())
        self.window.bind('<greater>', lambda event: self.delay_dec())
        self.window.bind('<less>', lambda event: self.delay_inc())
        self.window.bind('<space>', lambda event: self.toggle_running())
        self.window.bind('<Return>', lambda event: self.request_step())
        self.window.bind('<Shift-Return>', lambda event: self.request_round())
        self.window.createcommand('exit', self.quit)
        self.window.protocol("WM_DELETE_WINDOW", self.quit)

        if self.controller_socket:
            self.window.after_idle(self.request_initial)


    def init_mesh(self, game_state):
        width, height = game_state['shape']

        if self.geometry is None:
            screensize = (
                max(250, self.window.winfo_screenwidth() - 100),
                max(250, self.window.winfo_screenheight() - 100)
                )
        else:
            screensize = self.geometry
        scale_x = screensize[0] / width
        scale_y = screensize[1] / height
        scale = int(min(scale_x, scale_y, 50))

        self.mesh_graph = MeshGraph(width, height, scale * width, scale * height)
        self.init_bot_sprites(game_state['bots'])

    def update(self, game_state=None):
        # Update the times for the fps calculation (if we are running)
        # Our fps is only relevant for how often the bots update our viewer.
        # When the viewer updates itself, we do not count it.
        if self.running and game_state:
            self._times.append(time.monotonic())
            if len(self._times) > 3:
                # take the mean of the last two time differences
                # this could also be improved by taking up to four if available
                try:
                    self._fps = 2 / ((self._times[-1] - self._times[-2]) + (self._times[-2] - self._times[-3]))
                except ZeroDivisionError:
                    self._fps = 1 / 0.001
            else:
                self._fps = None
            if len(self._times) > 100:
                # Garbage collect old times
                self._times = self._times[-3:]

        if game_state is not None:
            self._game_state = game_state
        game_state = self._game_state
        if not game_state:
            return

        #if game_state['game_uuid'] != self.game_uuid:
        if not self.game_uuid:
            #self.game_uuid = game_state['game_uuid']
            self.game_uuid = 1
            self.init_mesh(game_state)

        if ((self.mesh_graph.screen_width, self.mesh_graph.screen_height)
            != (self.ui.game_canvas.winfo_width(), self.ui.game_canvas.winfo_height())):
            self.size_changed = True

        self.mesh_graph.screen_width = self.ui.game_canvas.winfo_width()
        self.mesh_graph.screen_height = self.ui.game_canvas.winfo_height()

        if self.mesh_graph.screen_width < 600:
            if self._default_font.cget('size') != 8:
                self._default_font.configure(size=8)
        else:
            if self._default_font.cget('size') != self._default_font_size:
                self._default_font.configure(size=self._default_font_size)

        self.draw_universe(game_state)

        eaten_food = []
        for food_pos, food_item in self.food_items.items():
            food_item.food_age = game_state['food_age'].get(food_pos, 0)
            if food_pos not in game_state["food"]:
                self.ui.game_canvas.delete(food_item.tag)
                eaten_food.append(food_pos)
        for food_pos in eaten_food:
            del self.food_items[food_pos]


        winning_team_idx = game_state.get("whowins")
        if winning_team_idx is None:
            self.draw_end_of_game(None)
        elif winning_team_idx in (0, 1):
            team_name = game_state["team_names"][winning_team_idx]
            self.draw_game_over(team_name)
        elif winning_team_idx == 2:
            self.draw_game_draw()

        self.size_changed = False

    def draw_universe(self, game_state):
        self.mesh_graph.num_x = game_state['shape'][0]
        self.mesh_graph.num_y = game_state['shape'][1]

        self.draw_grid()
        self.draw_selected(game_state)
        self.draw_line_of_sight(game_state)
        self.draw_bot_shadow(game_state)
        self.draw_background()
        self.draw_maze(game_state)
        self.draw_food(game_state)

        self.draw_title(game_state)
        self.draw_shadow_bots(game_state)
        self.draw_bots(game_state)

        self.draw_moves(game_state)

        self.draw_status_info(game_state)

    def draw_grid(self):
        """ Draws a light grid on the background.
        """
        if not self.size_changed:
            return
        self.ui.game_canvas.delete("grid")

        if not self._grid_enabled:
            return

        scale = self.mesh_graph.half_scale_x * 0.01

        def draw_line(x0, y0, x1, y1):
            x0_ = self.mesh_graph.mesh_to_screen_x(x0, 0)
            y0_ = self.mesh_graph.mesh_to_screen_y(y0, 0)
            x1_ = self.mesh_graph.mesh_to_screen_x(x1, 0)
            y1_ = self.mesh_graph.mesh_to_screen_y(y1, 0)
            self.ui.game_canvas.create_line(x0_, y0_, x1_, y1_, width=0.01, fill="#884488", tag="grid")

        for x in range(self.mesh_graph.mesh_width + 1):
            draw_line(x - 0.5, -0.5, x - 0.5, self.mesh_graph.mesh_height - 0.5)

        for y in range(self.mesh_graph.mesh_height + 1):
            draw_line(-0.5, y - 0.5, self.mesh_graph.mesh_width - 0.5, y - 0.5)

    def draw_line_of_sight(self, game_state):
        self.ui.game_canvas.delete("line_of_sight")
        if not self._grid_enabled:
            return

        scale = self.mesh_graph.half_scale_x * 0.1

        def draw_box(pos):
            ul = self.mesh_graph.mesh_to_screen(pos, (-1, -1))
            ur = self.mesh_graph.mesh_to_screen(pos, (1, -1))
            ll = self.mesh_graph.mesh_to_screen(pos, (-1, 1))
            lr = self.mesh_graph.mesh_to_screen(pos, (1, 1))

            self.ui.game_canvas.create_rectangle(*ul, *lr, width=2, outline='#111', tag=("line_of_sight",))

        bot = game_state['turn']
        if bot is None:
            # game has not started yet
            return

        line_col = STRONG_BLUE if bot % 2 == 0 else STRONG_RED
        fill_col = LIGHT_BLUE if bot % 2 == 0 else LIGHT_RED

        try:
            old_pos = tuple(game_state['requested_moves'][bot]['previous_position'])
        except TypeError:
            old_pos = game_state['bots'][bot]
        draw_box(old_pos)

        sight_distance = game_state["sight_distance"]
        # starting from old_pos, iterate over all positions that are up to sight_distance
        # steps away and put a border around the fields.
        border_cells_relative = set(
            (dx, dy)
            for dx in range(- sight_distance, sight_distance + 1)
            for dy in range(- sight_distance, sight_distance + 1)
            if abs(dx) + abs(dy) == sight_distance
        )

        def in_maze(x, y):
            return 0 <= x < game_state['shape'][0] and  0 <= y < game_state['shape'][1]

        def on_edge(x, y):
            return x == 0 or x == game_state['shape'][0] - 1 or y == 0 or y == game_state['shape'][1] - 1


        def draw_line(pos, line_col, loc):
            x0_ = self.mesh_graph.mesh_to_screen_x(pos[0], loc[0])
            y0_ = self.mesh_graph.mesh_to_screen_y(pos[1], loc[1])
            x1_ = self.mesh_graph.mesh_to_screen_x(pos[0], loc[2])
            y1_ = self.mesh_graph.mesh_to_screen_y(pos[1], loc[3])
            self.ui.game_canvas.create_line(x0_, y0_, x1_, y1_, width=3, fill=line_col, tag=("line_of_sight"))

        def draw_box(pos, fill_col):
            ul = self.mesh_graph.mesh_to_screen(pos, (-1, -1))
            ur = self.mesh_graph.mesh_to_screen(pos, (1, -1))
            ll = self.mesh_graph.mesh_to_screen(pos, (-1, 1))
            lr = self.mesh_graph.mesh_to_screen(pos, (1, 1))

            self.ui.game_canvas.create_rectangle(*ul, *lr, width=0, fill=fill_col, tag=("line_of_sight", "area_of_sight"))

        for dx in range(- sight_distance, sight_distance + 1):
            for dy in range(- sight_distance, sight_distance + 1):
                if abs(dx) + abs(dy) > sight_distance:
                    continue

                pos = (old_pos[0] + dx, old_pos[1] + dy)
                if not in_maze(pos[0], pos[1]):
                    continue

                # Currently not used
                # draw_box(pos, fill_col=fill_col)

                # add edge around cells at the line of sight max
                if (dx, dy) in border_cells_relative:
                    if dx >= 0:
                        draw_line(pos, loc=(1, 1, 1, -1), line_col=line_col)
                    if dx <= 0:
                        draw_line(pos, loc=(-1, 1, -1, -1), line_col=line_col)
                    if dy >= 0:
                        draw_line(pos, loc=(1, 1, -1, 1), line_col=line_col)
                    if dy <= 0:
                        draw_line(pos, loc=(1, -1, -1, -1), line_col=line_col)

                # add edge around cells at the edge of the maze
                if on_edge(pos[0], pos[1]):
                    if pos[0] == game_state['shape'][0] - 1:
                        draw_line(pos, loc=(1, 1, 1, -1), line_col=line_col)
                    if pos[0] == 0:
                        draw_line(pos, loc=(-1, 1, -1, -1), line_col=line_col)
                    if pos[1] == game_state['shape'][1] - 1:
                        draw_line(pos, loc=(1, 1, -1, 1), line_col=line_col)
                    if pos[1] == 0:
                        draw_line(pos, loc=(1, -1, -1, -1), line_col=line_col)

        self.ui.game_canvas.tag_lower("area_of_sight")
        self.ui.game_canvas.tag_raise("wall")


    def draw_bot_shadow(self, game_state):
        self.ui.game_canvas.delete("bot_shadow")
        if not self._grid_enabled:
            return

        border_col = "#000"
        fill_col = LIGHT_GREY

        scale = self.mesh_graph.half_scale_x * 0.1

        def draw_box(pos):
            ul = self.mesh_graph.mesh_to_screen(pos, (-1, -1))
            ur = self.mesh_graph.mesh_to_screen(pos, (1, -1))
            ll = self.mesh_graph.mesh_to_screen(pos, (-1, 1))
            lr = self.mesh_graph.mesh_to_screen(pos, (1, 1))

            self.ui.game_canvas.create_rectangle(*ul, *lr, width=2, outline='#111', tag=("bot_shadow",))

        bot = game_state['turn']
        if bot is None:
            # game has not started yet
            return

        try:
            old_pos = tuple(game_state['requested_moves'][bot]['previous_position'])
        except TypeError:
            old_pos = game_state['bots'][bot]

        boundary = game_state['shape'][0] / 2
        if bot % 2 == 0:
            in_homezone = old_pos[0] < boundary
        else:
            in_homezone = old_pos[0] >= boundary

        if not in_homezone:
            # We are a pacman. No shadow
            return

        draw_box(old_pos)

        sight_distance = game_state["shadow_distance"]
        # starting from old_pos, iterate over all positions that are up to sight_distance
        # steps away and put a border around the fields.
        border_cells_relative = set(
            (dx, dy)
            for dx in range(- sight_distance, sight_distance + 1)
            for dy in range(- sight_distance, sight_distance + 1)
            if abs(dx) + abs(dy) == sight_distance
        )

        def in_maze(x, y):
            return 0 <= x < game_state['shape'][0] and  0 <= y < game_state['shape'][1]

        def on_edge(x, y):
            return x == 0 or x == game_state['shape'][0] - 1 or y == 0 or y == game_state['shape'][1] - 1


        def draw_line(pos, line_col, loc):
            x0_ = self.mesh_graph.mesh_to_screen_x(pos[0], loc[0])
            y0_ = self.mesh_graph.mesh_to_screen_y(pos[1], loc[1])
            x1_ = self.mesh_graph.mesh_to_screen_x(pos[0], loc[2])
            y1_ = self.mesh_graph.mesh_to_screen_y(pos[1], loc[3])
            self.ui.game_canvas.create_line(x0_, y0_, x1_, y1_, width=2, fill=line_col, tag=("bot_shadow"))


        def draw_box(pos, fill_col):
            ul = self.mesh_graph.mesh_to_screen(pos, (-1, -1))
            ur = self.mesh_graph.mesh_to_screen(pos, (1, -1))
            ll = self.mesh_graph.mesh_to_screen(pos, (-1, 1))
            lr = self.mesh_graph.mesh_to_screen(pos, (1, 1))

            self.ui.game_canvas.create_rectangle(*ul, *lr, width=0, fill=fill_col, tag=("bot_shadow", "bot_shadow_area"))

        for dx in range(- sight_distance, sight_distance + 1):
            for dy in range(- sight_distance, sight_distance + 1):
                if abs(dx) + abs(dy) > sight_distance:
                    continue

                pos = (old_pos[0] + dx, old_pos[1] + dy)
                if not in_maze(pos[0], pos[1]):
                    continue

                draw_box(pos, fill_col=fill_col)

                # Border around the shadow removed for now
                #
                # # add edge around cells at the line of sight max
                # if (dx, dy) in border_cells_relative:
                #     if dx >= 0:
                #         draw_line(pos, loc=(1, 1, 1, -1), line_col=border_col)
                #     if dx <= 0:
                #         draw_line(pos, loc=(-1, 1, -1, -1), line_col=border_col)
                #     if dy >= 0:
                #         draw_line(pos, loc=(1, 1, -1, 1), line_col=border_col)
                #     if dy <= 0:
                #         draw_line(pos, loc=(1, -1, -1, -1), line_col=border_col)

                # # add edge around cells at the edge of the maze
                # if on_edge(pos[0], pos[1]):
                #     if pos[0] == game_state['shape'][0] - 1:
                #         draw_line(pos, loc=(1, 1, 1, -1), line_col=border_col)
                #     if pos[0] == 0:
                #         draw_line(pos, loc=(-1, 1, -1, -1), line_col=border_col)
                #     if pos[1] == game_state['shape'][1] - 1:
                #         draw_line(pos, loc=(1, 1, -1, 1), line_col=border_col)
                #     if pos[1] == 0:
                #         draw_line(pos, loc=(1, -1, -1, -1), line_col=border_col)

        self.ui.game_canvas.tag_lower("bot_shadow_area")
        self.ui.game_canvas.tag_lower("area_of_sight")
        self.ui.game_canvas.tag_raise("wall")


    def toggle_grid(self):
        self._grid_enabled = not self._grid_enabled
        self.size_changed = True
        self._check_grid_toggle_state()
        self.update()

    def toggle_fullscreen(self):
        self._fullscreen_enabled = not self._fullscreen_enabled
        self.size_changed = True
        if self._fullscreen_enabled:
            self.window.attributes('-fullscreen',True)
        else:
            self.window.attributes('-fullscreen',False)
        self.update()

    def _check_grid_toggle_state(self):
        if self._grid_enabled:
            self.ui.button_game_toggle_grid.config(text="debug")
        else:
            self.ui.button_game_toggle_grid.config(text="debug")

    def on_click(self, event):
        raw_x, raw_y = event.x, event.y
        selected = self.mesh_graph.screen_to_mesh_coord(event.x, event.y)
        if self.selected == selected:
            self.selected = None
        else:
            self.selected = selected
        self.update()

    def draw_background(self):
        """ Draws a line between blue and red team.
        """
        if not self.size_changed:
            return
        self.ui.game_canvas.delete("background")

        center = self.mesh_graph.screen_width // 2
        cols = (BLUE, RED, GREY)

        scale = self.mesh_graph.half_scale_x * 0.2

        for color, x_orig in zip(cols, (center - 3, center + 3, center)):
            y_top = self.mesh_graph.mesh_to_screen_y(0, 0)
            y_bottom = self.mesh_graph.mesh_to_screen_y(self.mesh_graph.mesh_height - 1, 0)
            self.ui.game_canvas.create_line(x_orig, y_top, x_orig, y_bottom, width=scale, fill=color, tag="background")

    def draw_title(self, game_state):
        self.ui.header_canvas.delete("title")

        center = self.ui.header_canvas.winfo_width() // 2

        try:
            team_time = game_state["team_time"]
        except (KeyError, TypeError):
            team_time = [0, 0]

        left_name = game_state["team_names"][0]
        right_name = game_state["team_names"][1]

        left_score = game_state["score"][0]
        right_score = game_state["score"][1]

        left_info = game_state["team_infos"][0]
        right_info = game_state["team_infos"][1]

        left_info = f"({left_info})" if left_info else ""
        right_info = f"({right_info})" if right_info else ""

        left_team = f"{left_info} {left_name} {left_score} "
        right_team = f" {right_score} {right_name} {right_info}"

        font_size = guess_size(left_team + ' : ' + right_team,
                               self.ui.header_canvas.winfo_width(),
                               30,
                               rel_size = 1)

        def status(team_idx):
            try:
                # in case we had a fatal error, do not print the regular status
                if len(game_state['fatal_errors'][team_idx]) > 0:
                    # TODO: We only print the first fatal error (it is a list of errors) for now
                    err_type = game_state['fatal_errors'][team_idx][0]['type']
                    err_desc = game_state['fatal_errors'][team_idx][0]['description']
                    ret = f"FATAL: {err_type} {err_desc}"
                else:
                    # sum the deaths of both bots in this team
                    deaths = game_state['deaths'][team_idx] + game_state['deaths'][team_idx+2]
                    kills = game_state['kills'][team_idx] + game_state['kills'][team_idx+2]
                    ret = "Errors: %d, Kills: %d, Deaths: %d, Time: %.2f" % (game_state["num_errors"][team_idx], kills, deaths, game_state["team_time"][team_idx])
                return ret
            except TypeError:
                return ""

        left_status = status(0)
        right_status = status(1)
        status_font_size = max(font_size - 5, 3)

        top = 15
        spacer = 3

        middle_colon = self.ui.header_canvas.create_text(center, top, text=":", font=(None, font_size), tag="title", anchor=tkinter.CENTER)
        middle_colon_bottom = self.ui.header_canvas.bbox(middle_colon)[3]
        spacer = (self.ui.header_canvas.bbox(middle_colon)[3] - self.ui.header_canvas.bbox(middle_colon)[1]) / 2

        self.ui.header_canvas.create_text(center, top, text=left_team, font=(None, font_size), fill=BLUE, tag="title", anchor=tkinter.E)
        self.ui.header_canvas.create_text(center+2, top, text=right_team, font=(None, font_size), fill=RED, tag="title", anchor=tkinter.W)

        bottom_text = self.ui.header_canvas.create_text(0 + 5, 20 + font_size, text=" " + left_status, font=(None, status_font_size), tag="title", anchor=tkinter.W)
        self.ui.header_canvas.create_text(self.ui.header_canvas.winfo_width() - 5, 20 + font_size, text=right_status + " ", font=(None, status_font_size), tag="title", anchor=tkinter.E)

    def draw_status_info(self, game_state):
        round = "–" if game_state["round"] is None else game_state["round"]
        max_rounds = "–" if game_state["max_rounds"] is None else game_state["max_rounds"]
        turn = "–" if game_state["turn"] is None else game_state["turn"]
        layout_name = "–" if game_state["layout_name"] is None else game_state["layout_name"]

        round_info = f"Round {round:>3}/{max_rounds}"
        self.ui.status_round_info.config(text=round_info)

        if self._fps is not None:
            fps_info = "%.f fps" % self._fps
        else:
            fps_info = "– fps"
        self.ui.status_fps_info.config(text=fps_info)

        bot_colors = [BLUE, RED, BLUE, RED]
        for idx, circle in enumerate(self.ui.bot_traffic_lights_c):
            if turn == idx:
                self.ui.bot_traffic_lights_canvas.itemconfig(circle, fill=bot_colors[idx])
            else:
                self.ui.bot_traffic_lights_canvas.itemconfig(circle, fill="#bbb")

        self.ui.status_layout_info.config(text=layout_name)

    def draw_selected(self, game_state):
        self.ui.game_canvas.delete("selected")
        if self.selected:
            def field_status(pos):
                has_food = pos in game_state['food']
                is_wall = pos in game_state['walls']
                bots = [idx for idx, bot in enumerate(game_state['bots']) if bot==pos]
                if pos[0] < (game_state['shape'][0] // 2):
                    zone = "blue"
                else:
                    zone = "red"

                if is_wall:
                    contents = ["wall"]
                elif has_food:
                    contents = ["food"]
                else:
                    contents = []

                if bots:
                    bot_map = {0:'blue 0', 1:'red 0', 2:'blue 1', 3:'red 1'}
                    contents += ["bots(" \
                                 + ", ".join(bot_map[bot] for bot in bots) \
                                 + ")"]

                contents = " ".join(contents)
                if not contents:
                    contents = "empty"

                return f"{pos} in {zone} zone: {contents}"

            self.ui.status_selected.config(text=field_status(self.selected))

            ul = self.mesh_graph.mesh_to_screen(self.selected, (-1, -1))
            ur = self.mesh_graph.mesh_to_screen(self.selected, (1, -1))
            ll = self.mesh_graph.mesh_to_screen(self.selected, (-1, 1))
            lr = self.mesh_graph.mesh_to_screen(self.selected, (1, 1))

            self.ui.game_canvas.create_rectangle(*ul, *lr, fill=SELECTED, tag=("selected",))
            self.ui.game_canvas.tag_lower("selected")
        else:
            self.ui.status_selected.config(text="nothing selected")


    def draw_end_of_game(self, display_string):
        """ Draw an end of game string. """
        self.ui.game_canvas.delete("gameover")

        if display_string is None:
            return

        center = (self.mesh_graph.screen_width // 2,
                  self.mesh_graph.screen_height //2)

        font_size = guess_size(display_string,
                               self.mesh_graph.screen_width,
                               self.mesh_graph.screen_height,
                               rel_size = +1)

        for i in [-2, -1, 0, 1, 2]:
            for j in [-2, -1, 0, 1, 2]:
                self.ui.game_canvas.create_text(center[0] - i, center[1] - j,
                        text=display_string,
                        font=(None, font_size, "bold"),
                        fill="#ED1B22", tag="gameover",
                        justify=tkinter.CENTER, anchor=tkinter.CENTER)

        self.ui.game_canvas.create_text(center[0] , center[1] ,
                text=display_string,
                font=(None, font_size, "bold"),
                fill="#FFC903", tag="gameover",
                justify=tkinter.CENTER, anchor=tkinter.CENTER)


    def draw_game_over(self, win_name):
        """ Draw the game over string. """
        # shorten the winning name
        plural = '' if win_name.endswith('s') else 's'
        if len(win_name) > 25:
            win_name = win_name[:22] + '...'
        self.draw_end_of_game("GAME OVER\n%s win%s!" % (win_name, plural))

    def draw_game_draw(self):
        """ Draw the game draw string. """
        self.draw_end_of_game("GAME OVER\nDRAW!")

    def clear(self):
        self.ui.game_canvas.delete(tkinter.ALL)

    def draw_food(self, game_state):
#        if not self.size_changed:
#            return
        self.ui.game_canvas.delete("food")
        self.food_items = {}
        max_food_age = game_state["max_food_age"]
        for position in game_state['food']:
            model_x, model_y = position
            food_age = game_state['food_age'].get(position, 0)
            food_item = Food(
                self.mesh_graph,
                position=(model_x, model_y),
                food_age=food_age,
                max_food_age=max_food_age,
            )
            food_item.draw(self.ui.game_canvas, show_lifetime=False)
            self.food_items[position] = food_item

    def draw_maze(self, game_state):
        if not self.size_changed:
            return
        self.ui.game_canvas.delete("wall")
        # we keep all wall items stored in a list
        # some versions of Python seem to forget about drawing
        # them otherwise
        self.wall_items = []
        num = 0
        for wall in game_state['walls']:
            model_x, model_y = wall
            wall_neighbors = [(dx, dy)
                              for dx in [-1, 0, 1]
                              for dy in [-1, 0, 1]
                              if (model_x + dx, model_y + dy) in game_state['walls']]
            wall_item = Wall(self.mesh_graph, wall_neighbors=wall_neighbors, position=(model_x, model_y))
            wall_item.draw(self.ui.game_canvas)
            self.wall_items.append(wall_item)
            num += 1

    def init_bot_sprites(self, bot_positions):
        for sprite in self.bot_sprites.values():
            sprite.delete(self.ui.game_canvas)
        self.bot_sprites = {
            idx: BotSprite(self.mesh_graph, team=idx % 2, bot_id=idx, position=bot)
            for idx, bot in enumerate(bot_positions)
        }
        for sprite in self.shadow_sprites.values():
            sprite.delete(self.ui.game_canvas)
        self.shadow_sprites = {
            idx: BotSprite(self.mesh_graph, team=idx % 2, bot_id=idx, position=None, shadow=True)
            for idx, bot in enumerate(bot_positions)
        }

    def draw_moves(self, game_state):
        if game_state['turn'] is None:
            return

        self.ui.game_canvas.delete("arrow")
        # we keep all arrow items stored in a list
        # some versions of Python seem to forget about drawing
        # them otherwise
        self.arrow_items = []

        if not self._grid_enabled:
            return

        bot = game_state['turn']
        try:
            old_pos = tuple(game_state['requested_moves'][bot]['previous_position'])
        except TypeError:
            old_pos = None
        try:
            req_pos = tuple(game_state['requested_moves'][bot]['requested_position'])
        except TypeError:
            req_pos = None

        if game_state['requested_moves'][bot]['success'] and tuple(game_state['bots'][bot]) != tuple(game_state['requested_moves'][bot]['requested_position']):
            # Bot has committed suicide. Show two arrows.
            arrow_item1 = Arrow(self.mesh_graph,
                            position=old_pos,
                            req_pos=game_state['requested_moves'][bot]['requested_position'],
                            success=True,
                            head=False)
            arrow_item1.draw(self.ui.game_canvas)

            arrow_item2 = Arrow(self.mesh_graph,
                            position=game_state['requested_moves'][bot]['requested_position'],
                            req_pos=game_state['bots'][bot],
                            success=True)
            arrow_item2.draw(self.ui.game_canvas)

            self.arrow_items.append(arrow_item1)
            self.arrow_items.append(arrow_item2)
        else:
            arrow_item = Arrow(self.mesh_graph,
                            position=old_pos,
                            req_pos=game_state['bots'][bot],
                            success=game_state['requested_moves'][bot]['success'])
            arrow_item.draw(self.ui.game_canvas)
            self.arrow_items.append(arrow_item)


    def draw_bots(self, game_state):
        if game_state:
            for bot_id, was_killed in enumerate(game_state["bot_was_killed"]):
                if was_killed:
                    self.bot_sprites[bot_id].position = None

        for bot_id, bot_sprite in self.bot_sprites.items():
            say = game_state and game_state["say"][bot_id]
            bot_sprite.move_to(game_state["bots"][bot_sprite.bot_id],
                               self.ui.game_canvas,
                               game_state,
                               force=self.size_changed,
                               say=say,
                               show_id=self._grid_enabled)

    def draw_shadow_bots(self, game_state):
        # Draw the shadowbots when debug mode (grid) is enabled
        # Otherwise make sure that they are deleted

        if game_state['turn'] is None:
            # We cannot show shadow bots before the first turn has been played,
            # as we would only see the bots from the set_initial phase, and
            # only the blue bots (those that were shown to the second, red, player).
            # Given that this information is not available to the client API,
            # we must hide the shadow bots here.
            # TODO: This should be fixed inside game.py
            return

        for bot_id, bot_sprite in self.shadow_sprites.items():
            if self._grid_enabled:
                shadow_bots = game_state.get('noisy_positions')
            else:
                shadow_bots = None

            if shadow_bots is None or shadow_bots[bot_id] is None:
                bot_sprite.delete(self.ui.game_canvas)
            else:
                bot_sprite.move_to(shadow_bots[bot_id],
                                    self.ui.game_canvas,
                                    game_state,
                                    force=self.size_changed,
                                    show_id=self._grid_enabled)

    def toggle_running(self):
        # We change from running to stopping or the other way round
        # Clean up the times for fps calculation as they will be wrong
        self._fps = None
        self._times = []

        self.running = not self.running
        if self.running:
            self.request_step()

    def request_initial(self):
        if self.controller_socket:
            _logger.debug('---> set_initial')
            self.controller_socket.send_json({"__action__": "set_initial"})

    def request_step(self):
        if not self.controller_socket:
            return

        if self._game_state['gameover']:
            return

        if self._stop_after is not None:
            next_step = next_round_turn(self._game_state)
            if (next_step['round'] < self._stop_after):
                _logger.debug('---> play_step')
                self.controller_socket.send_json({"__action__": "play_step"})
            else:
                self._stop_after = None
                self.running = False
                self._delay = self._stop_after_delay
        else:
            _logger.debug('---> play_step')
            self.controller_socket.send_json({"__action__": "play_step"})

    def request_round(self):
        if not self.controller_socket:
            return

        if self._game_state['gameover']:
            return

        if self._game_state['round'] is not None:
            next_step = next_round_turn(self._game_state)
            self._stop_after = next_step['round'] + 1
        else:
            self._stop_after = 1
            self._delay = self._min_delay
        self.request_step()

    def observe(self, game_state):
        step = (game_state['round'], game_state['turn'])
        if step in self._observed_steps:
            skip_request = True
        else:
            skip_request = False
            self._observed_steps.add(step)
        # ensure walls, foods and bots positions are list of tuples
        game_state['walls'] = _ensure_list_tuples(game_state['walls'])
        game_state['food'] = _ensure_list_tuples(game_state['food'])
        game_state['bots'] = _ensure_list_tuples(game_state['bots'])
        game_state['shape'] = tuple(game_state['shape'])
        game_state['food_age'] = {tuple(pos): food_age for pos, food_age in game_state['food_age']}
        self.update(game_state)
        if self._stop_after is not None:
            if self._stop_after == 0:
                self._stop_after = None
                self.running = False
                self._delay = self._stop_after_delay
            else:
                if skip_request:
                    _logger.debug("Skipping next request.")
                else:
                    self.window.after(self._delay, self.request_step)
        elif self.running:
            if skip_request:
                _logger.debug("Skipping next request.")
            else:
                self.window.after(self._delay, self.request_step)


    def on_quit(self):
        """ override for things which must be done when we exit.
        """
        self.running = False
        if self.controller_socket:
            _logger.debug('---> exit')
            self.controller_socket.send_json({"__action__": "exit"})
        else:
            # force closing the window (though this might not work)
            wm_delete_window_handler()

    def quit(self):
        self.on_quit()
        self.window.quit()

    def delay_inc(self):
        self._delay += 5
        self._check_speed_button_state()

    def delay_dec(self):
        # Tk may break if self._delay is lower than zero.
        # (For some systems a value < 1 is already too fast.)
        self._delay = max(self._delay - 5, self._min_delay)
        self._check_speed_button_state()

    def _check_speed_button_state(self):
        try:
            # self.ui.button_game_speed_faster
            # may not be available yet (or may be None).
            # If this is the case, we’ll do nothing at all.
            if self._delay <= self._min_delay:
                self.ui.button_game_speed_faster.config(state=tkinter.DISABLED)
            else:
                self.ui.button_game_speed_faster.config(state=tkinter.NORMAL)
        except AttributeError:
            pass



""" Base classes for player implementations. """


def speaking_player(bot, state):
    """ A player that makes moves at random and tells us about it. """
    move = bot.random.choice(bot.legal_positions)
    bot.say(f"Going {move}.")
    return move


def stepping_player(*bot_moves):
    """ A Player with predetermined set of moves.

    Parameters
    ----------
    moves : list of moves or str of shorthand symbols
        the moves to make in order, see notes below

    Notes
    -----
    The ``moves`` argument can either be a list of moves, e.g. ``[west, east,
    south, north, stop]`` or a string of shorthand symbols, where the equivalent
    of the previous example is: ``'><v^-'``.

    """
    _MOVES = {
        '^': (0, -1),
        'v': (0, 1),
        '<': (-1, 0),
        '>': (1, 0),
        '-': (0, 0),
    }

    def convert(next_move):
        if isinstance(next_move, str):
            return _MOVES[next_move]
        else:
            return next_move

    converted_moves = [map(convert, m) for m in bot_moves]
    iterators = [iter(m) for m in converted_moves]

    def move(bot, state):
        moves_iter = iterators[bot.turn]
        try:
            next_move = next(moves_iter)
            return (bot.position[0] + next_move[0], bot.position[1] + next_move[1])
        except StopIteration:
            raise ValueError()
    return move


def round_based_player(*bot_moves):
    """ A Player which makes a decision dependent on the round index
    in a dict or list. (Or anything which responds to moves[idx].)

    Parameters
    ----------
    moves : list or dict of moves
        the moves to make, a move is determined by moves[round]
    """
    def move(bot, state):
        try:
            next_move = bot_moves[bot.turn][bot.round]
            return (bot.position[0] + next_move[0], bot.position[1] + next_move[1])
        except (IndexError, KeyError):
            return bot.position
    return move


def move_exception_player(bot, state):
    """ Player that raises an Exception on get_move(). """
    raise Exception("Exception from MoveExceptionPlayer.")


def debuggable_player(bot, state):
    """ Player which invokes pdb on each move.

    Setting ``positions`` inside the debugger will change
    its behaviour.
    """
    position = bot.position
    breakpoint()
    return position
